@@ -107,7 +107,7 @@ created.
 # The issues related to nested repos and subrepos are probably not fundamental
 # ones. Patches to fix them are welcome.
 
-from __future__ import absolute_import
+from __future__ import annotations
 
 import codecs
 import os
@@ -118,7 +118,7 @@ import weakref
 
 from mercurial.i18n import _
 from mercurial.node import hex
-from mercurial.pycompat import open
+
 from mercurial import (
     context,
     encoding,
@@ -132,6 +132,9 @@ from mercurial import (
     scmutil,
     util,
 )
+
+# no-check-code because we're accessing private information only public in pure
+from mercurial.pure import parsers
 from mercurial import match as matchmod
 from mercurial.utils import (
     hashutil,
@@ -329,14 +332,26 @@ def overridewalk(orig, self, match, subrepos, unknown, ignored, full=True):
     matchfn = match.matchfn
     matchalways = match.always()
     dmap = self._map
-    if util.safehasattr(dmap, b'_map'):
+    if hasattr(dmap, '_map'):
         # for better performance, directly access the inner dirstate map if the
         # standard dirstate implementation is in use.
         dmap = dmap._map
+
+    has_mtime = parsers.DIRSTATE_V2_HAS_MTIME
+    mtime_is_ambiguous = parsers.DIRSTATE_V2_MTIME_SECOND_AMBIGUOUS
+    mask = has_mtime | mtime_is_ambiguous
+
+    # All entries that may not be clean
     nonnormalset = {
         f
         for f, e in self._map.items()
-        if e.v1_state() != "n" or e.v1_mtime() == -1
+        if not e.maybe_clean
+        # same as "not has_time or has_ambiguous_time", but factored to only
+        # need a single access to flags for performance.
+        # `mask` removes all irrelevant bits, then we flip the `mtime` bit so
+        # its `true` value is NOT having a mtime, then check if either bit
+        # is set.
+        or bool((e.v2_data()[0] & mask) ^ has_mtime)
     }
 
     copymap = self._map.copymap
@@ -502,15 +517,11 @@ def overridewalk(orig, self, match, subrepos, unknown, ignored, full=True):
             visit.update(f for f in copymap if f not in results and matchfn(f))
     else:
         if matchalways:
-            visit.update(
-                f for f, st in pycompat.iteritems(dmap) if f not in results
-            )
+            visit.update(f for f, st in dmap.items() if f not in results)
             visit.update(f for f in copymap if f not in results)
         else:
             visit.update(
-                f
-                for f, st in pycompat.iteritems(dmap)
-                if f not in results and matchfn(f)
+                f for f, st in dmap.items() if f not in results and matchfn(f)
             )
             visit.update(f for f in copymap if f not in results and matchfn(f))
 
@@ -552,11 +563,11 @@ def overridestatus(
         try:
             if b'FSMONITOR_LOG_FILE' in encoding.environ:
                 fn = encoding.environ[b'FSMONITOR_LOG_FILE']
-                f = open(fn, b'wb')
+                f = open(fn, 'wb')
             else:
                 fn = b'fsmonitorfail.log'
                 f = self.vfs.open(fn, b'wb')
-        except (IOError, OSError):
+        except OSError:
             self.ui.warn(_(b'warning: unable to write to %s\n') % fn)
             return
 
@@ -656,22 +667,24 @@ def overridestatus(
         quiet = self.ui.quiet
         self.ui.quiet = True
         fout, ferr = self.ui.fout, self.ui.ferr
-        self.ui.fout = self.ui.ferr = open(os.devnull, b'wb')
 
-        try:
-            rv2 = orig(
-                node1,
-                node2,
-                match,
-                listignored,
-                listclean,
-                listunknown,
-                listsubrepos,
-            )
-        finally:
-            self.dirstate._fsmonitordisable = False
-            self.ui.quiet = quiet
-            self.ui.fout, self.ui.ferr = fout, ferr
+        with open(os.devnull, 'wb') as fp:
+            self.ui.fout = self.ui.ferr = fp
+
+            try:
+                rv2 = orig(
+                    node1,
+                    node2,
+                    match,
+                    listignored,
+                    listclean,
+                    listunknown,
+                    listsubrepos,
+                )
+            finally:
+                self.dirstate._fsmonitordisable = False
+                self.ui.quiet = quiet
+                self.ui.fout, self.ui.ferr = fout, ferr
 
         # clean isn't tested since it's set to True above
         with self.wlock():
@@ -686,7 +699,7 @@ def overridestatus(
     )
 
 
-class poststatus(object):
+class poststatus:
     def __init__(self, startclock):
         self._startclock = pycompat.sysbytes(startclock)
 
@@ -713,18 +726,18 @@ def makedirstate(repo, dirstate):
             self._repo = weakref.proxy(repo)
 
         def walk(self, *args, **kwargs):
-            orig = super(fsmonitordirstate, self).walk
+            orig = super().walk
             if self._fsmonitordisable:
                 return orig(*args, **kwargs)
             return overridewalk(orig, self, *args, **kwargs)
 
         def rebuild(self, *args, **kwargs):
             self._fsmonitorstate.invalidate()
-            return super(fsmonitordirstate, self).rebuild(*args, **kwargs)
+            return super().rebuild(*args, **kwargs)
 
         def invalidate(self, *args, **kwargs):
             self._fsmonitorstate.invalidate()
-            return super(fsmonitordirstate, self).invalidate(*args, **kwargs)
+            return super().invalidate(*args, **kwargs)
 
     dirstate.__class__ = fsmonitordirstate
     dirstate._fsmonitorinit(repo)
@@ -733,7 +746,7 @@ def makedirstate(repo, dirstate):
 def wrapdirstate(orig, self):
     ds = orig(self)
     # only override the dirstate when Watchman is available for the repo
-    if util.safehasattr(self, b'_fsmonitorstate'):
+    if hasattr(self, '_fsmonitorstate'):
         makedirstate(self, ds)
     return ds
 
@@ -744,9 +757,10 @@ def extsetup(ui):
     )
     if pycompat.isdarwin:
         # An assist for avoiding the dangling-symlink fsevents bug
-        extensions.wrapfunction(os, b'symlink', wrapsymlink)
+        extensions.wrapfunction(os, 'symlink', wrapsymlink)
 
-    extensions.wrapfunction(merge, b'_update', wrapupdate)
+    extensions.wrapfunction(merge, '_update', wrapupdate)
+    merge._fs_monitor_loaded = True
 
 
 def wrapsymlink(orig, source, link_name):
@@ -761,7 +775,7 @@ def wrapsymlink(orig, source, link_name):
             pass
 
 
-class state_update(object):
+class state_update:
     """This context manager is responsible for dispatching the state-enter
     and state-leave signals to the watchman service. The enter and leave
     methods can be invoked manually (for scenarios where context manager
@@ -800,7 +814,7 @@ class state_update(object):
             self.oldnode = self.repo[b'.'].node()
 
         if self.repo.currentwlock() is None:
-            if util.safehasattr(self.repo, b'wlocknostateupdate'):
+            if hasattr(self.repo, 'wlocknostateupdate'):
                 self._lock = self.repo.wlocknostateupdate()
             else:
                 self._lock = self.repo.wlock()
@@ -828,7 +842,7 @@ class state_update(object):
                 self._lock.release()
 
     def _state(self, cmd, commithash, status=b'ok'):
-        if not util.safehasattr(self.repo, b'_watchmanclient'):
+        if not hasattr(self.repo, '_watchmanclient'):
             return False
         try:
             self.repo._watchmanclient.command(
@@ -880,9 +894,8 @@ def wrapupdate(
     mergeancestor=False,
     labels=None,
     matcher=None,
-    **kwargs
+    **kwargs,
 ):
-
     distance = 0
     partial = True
     oldnode = repo[b'.'].node()
@@ -908,7 +921,7 @@ def wrapupdate(
             mergeancestor,
             labels,
             matcher,
-            **kwargs
+            **kwargs,
         )
 
 
@@ -967,14 +980,14 @@ def reposetup(ui, repo):
 
         class fsmonitorrepo(repo.__class__):
             def status(self, *args, **kwargs):
-                orig = super(fsmonitorrepo, self).status
+                orig = super().status
                 return overridestatus(orig, self, *args, **kwargs)
 
             def wlocknostateupdate(self, *args, **kwargs):
-                return super(fsmonitorrepo, self).wlock(*args, **kwargs)
+                return super().wlock(*args, **kwargs)
 
             def wlock(self, *args, **kwargs):
-                l = super(fsmonitorrepo, self).wlock(*args, **kwargs)
+                l = super().wlock(*args, **kwargs)
                 if not ui.configbool(
                     b"experimental", b"fsmonitor.transaction_notify"
                 ):
