@@ -5,16 +5,13 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-from __future__ import absolute_import
+from __future__ import annotations
 
 import hashlib
+import typing
 
 from mercurial.i18n import _
-from mercurial.node import bin, hex, nullid, short
-from mercurial.pycompat import (
-    getattr,
-    setattr,
-)
+from mercurial.node import bin, hex, short
 
 from mercurial import (
     bundle2,
@@ -24,11 +21,9 @@ from mercurial import (
     error,
     exchange,
     exthelper,
-    localrepo,
-    pycompat,
+    requirements as req_mod,
     revlog,
     scmutil,
-    util,
     vfs as vfsmod,
     wireprotov1server,
 )
@@ -38,8 +33,6 @@ from mercurial.upgrade_utils import (
     engine as upgrade_engine,
 )
 
-from mercurial.interfaces import repository
-
 from mercurial.utils import (
     storageutil,
     stringutil,
@@ -52,34 +45,38 @@ from . import (
     pointer,
 )
 
+if typing.TYPE_CHECKING:
+    from typing import (
+        Callable,
+    )
+    from mercurial.interfaces.types import (
+        RepoT,
+        RequirementSetT,
+        UiT,
+    )
+
 eh = exthelper.exthelper()
 
 
-@eh.wrapfunction(localrepo, b'makefilestorage')
-def localrepomakefilestorage(orig, requirements, features, **kwargs):
-    if b'lfs' in requirements:
-        features.add(repository.REPO_FEATURE_LFS)
-
-    return orig(requirements=requirements, features=features, **kwargs)
-
-
-@eh.wrapfunction(changegroup, b'allsupportedversions')
-def allsupportedversions(orig, ui):
-    versions = orig(ui)
+@eh.wrapfunction(changegroup, 'allsupportedversions')
+def allsupportedversions(
+    orig: Callable[[RepoT], set[bytes]], repo: RepoT
+) -> set[bytes]:
+    versions = orig(repo)
     versions.add(b'03')
     return versions
 
 
-@eh.wrapfunction(wireprotov1server, b'_capabilities')
+@eh.wrapfunction(wireprotov1server, '_capabilities')
 def _capabilities(orig, repo, proto):
     '''Wrap server command to announce lfs server capability'''
     caps = orig(repo, proto)
-    if util.safehasattr(repo.svfs, b'lfslocalblobstore'):
+    if hasattr(repo.svfs, 'lfslocalblobstore'):
         # Advertise a slightly different capability when lfs is *required*, so
         # that the client knows it MUST load the extension.  If lfs is not
         # required on the server, there's no reason to autoload the extension
         # on the client.
-        if b'lfs' in repo.requirements:
+        if req_mod.LFS_REQUIREMENT in repo.requirements:
             caps.append(b'lfs-serve')
 
         caps.append(b'lfs')
@@ -116,10 +113,10 @@ def readfromstore(self, text):
     if hgmeta or text.startswith(b'\1\n'):
         text = storageutil.packmeta(hgmeta, text)
 
-    return (text, True, {})
+    return (text, True)
 
 
-def writetostore(self, text, sidedata):
+def writetostore(self, text):
     # hg filelog metadata (includes rename, etc)
     hgmeta, offset = storageutil.parsemeta(text)
     if offset and offset > 0:
@@ -143,7 +140,7 @@ def writetostore(self, text, sidedata):
 
     # translate hg filelog metadata to lfs metadata with "x-hg-" prefix
     if hgmeta is not None:
-        for k, v in pycompat.iteritems(hgmeta):
+        for k, v in hgmeta.items():
             metadata[b'x-hg-%s' % k] = v
 
     rawtext = metadata.serialize()
@@ -158,7 +155,7 @@ def _islfs(rlog, node=None, rev=None):
         rev = rlog.rev(node)
     else:
         node = rlog.node(rev)
-    if node == nullid:
+    if node == rlog.nullid:
         return False
     flags = rlog.flags(rev)
     return bool(flags & revlog.REVIDX_EXTSTORED)
@@ -176,7 +173,7 @@ def filelogaddrevision(
     cachedelta=None,
     node=None,
     flags=revlog.REVIDX_DEFAULT_FLAGS,
-    **kwds
+    **kwds,
 ):
     # The matcher isn't available if reposetup() wasn't called.
     lfstrack = self._revlog.opener.options.get(b'lfstrack')
@@ -201,7 +198,7 @@ def filelogaddrevision(
         cachedelta=cachedelta,
         node=node,
         flags=flags,
-        **kwds
+        **kwds,
     )
 
 
@@ -219,6 +216,20 @@ def filelogrenamed(orig, self, node):
     return orig(self, node)
 
 
+# Wrapping may also be applied by remotefilelog
+def filelog_has_meta(orig, self, node):
+    if _islfs(self._revlog, node):
+        rawtext = self._revlog.rawdata(node)
+        if not rawtext:
+            return False
+        metadata = pointer.deserialize(rawtext)
+        for k in metadata:
+            if k.startswith(b'x-hg-'):
+                return True
+        return False
+    return orig(self, node)
+
+
 # Wrapping may also be applied by remotefilelog
 def filelogsize(orig, self, rev):
     if _islfs(self._revlog, rev=rev):
@@ -229,7 +240,7 @@ def filelogsize(orig, self, rev):
     return orig(self, rev)
 
 
-@eh.wrapfunction(revlog, b'_verify_revision')
+@eh.wrapfunction(revlog, '_verify_revision')
 def _verify_revision(orig, rl, skipflags, state, node):
     if _islfs(rl, node=node):
         rawtext = rl.rawdata(node)
@@ -248,7 +259,7 @@ def _verify_revision(orig, rl, skipflags, state, node):
     orig(rl, skipflags, state, node)
 
 
-@eh.wrapfunction(context.basefilectx, b'cmp')
+@eh.wrapfunction(context.basefilectx, 'cmp')
 def filectxcmp(orig, self, fctx):
     """returns True if text is different than fctx"""
     # some fctx (ex. hg-git) is not based on basefilectx and do not have islfs
@@ -260,7 +271,7 @@ def filectxcmp(orig, self, fctx):
     return orig(self, fctx)
 
 
-@eh.wrapfunction(context.basefilectx, b'isbinary')
+@eh.wrapfunction(context.basefilectx, 'isbinary')
 def filectxisbinary(orig, self):
     if self.islfs():
         # fast path: use lfs metadata to answer isbinary
@@ -274,13 +285,13 @@ def filectxislfs(self):
     return _islfs(self.filelog()._revlog, self.filenode())
 
 
-@eh.wrapfunction(cmdutil, b'_updatecatformatter')
+@eh.wrapfunction(cmdutil, '_updatecatformatter')
 def _updatecatformatter(orig, fm, ctx, matcher, path, decode):
     orig(fm, ctx, matcher, path, decode)
     fm.data(rawdata=ctx[path].rawdata())
 
 
-@eh.wrapfunction(scmutil, b'wrapconvertsink')
+@eh.wrapfunction(scmutil, 'wrapconvertsink')
 def convertsink(orig, sink):
     sink = orig(sink)
     if sink.repotype == b'hg':
@@ -297,7 +308,7 @@ def convertsink(orig, sink):
                 full,
                 cleanp2,
             ):
-                pc = super(lfssink, self).putcommit
+                pc = super().putcommit
                 node = pc(
                     files,
                     copies,
@@ -309,13 +320,13 @@ def convertsink(orig, sink):
                     cleanp2,
                 )
 
-                if b'lfs' not in self.repo.requirements:
+                if req_mod.LFS_REQUIREMENT not in self.repo.requirements:
                     ctx = self.repo[node]
 
                     # The file list may contain removed files, so check for
                     # membership before assuming it is in the context.
                     if any(f in ctx and ctx[f].islfs() for f, n in files):
-                        self.repo.requirements.add(b'lfs')
+                        self.repo.requirements.add(req_mod.LFS_REQUIREMENT)
                         scmutil.writereporequirements(self.repo)
 
                 return node
@@ -327,7 +338,7 @@ def convertsink(orig, sink):
 
 # bundlerepo uses "vfsmod.readonlyvfs(othervfs)", we need to make sure lfs
 # options and blob stores are passed from othervfs to the new readonlyvfs.
-@eh.wrapfunction(vfsmod.readonlyvfs, b'__init__')
+@eh.wrapfunction(vfsmod.readonlyvfs, '__init__')
 def vfsinit(orig, self, othervfs):
     orig(self, othervfs)
     # copy lfs related options
@@ -336,15 +347,15 @@ def vfsinit(orig, self, othervfs):
             self.options[k] = v
     # also copy lfs blobstores. note: this can run before reposetup, so lfs
     # blobstore attributes are not always ready at this time.
-    for name in [b'lfslocalblobstore', b'lfsremoteblobstore']:
-        if util.safehasattr(othervfs, name):
+    for name in ['lfslocalblobstore', 'lfsremoteblobstore']:
+        if hasattr(othervfs, name):
             setattr(self, name, getattr(othervfs, name))
 
 
 def _prefetchfiles(repo, revmatches):
     """Ensure that required LFS blobs are present, fetching them as a group if
     needed."""
-    if not util.safehasattr(repo.svfs, b'lfslocalblobstore'):
+    if not hasattr(repo.svfs, 'lfslocalblobstore'):
         return
 
     pointers = []
@@ -368,7 +379,7 @@ def _prefetchfiles(repo, revmatches):
 
 def _canskipupload(repo):
     # Skip if this hasn't been passed to reposetup()
-    if not util.safehasattr(repo.svfs, b'lfsremoteblobstore'):
+    if not hasattr(repo.svfs, 'lfsremoteblobstore'):
         return True
 
     # if remotestore is a null store, upload is a no-op and can be skipped
@@ -377,7 +388,7 @@ def _canskipupload(repo):
 
 def candownload(repo):
     # Skip if this hasn't been passed to reposetup()
-    if not util.safehasattr(repo.svfs, b'lfsremoteblobstore'):
+    if not hasattr(repo.svfs, 'lfsremoteblobstore'):
         return False
 
     # if remotestore is a null store, downloads will lead to nothing
@@ -385,10 +396,7 @@ def candownload(repo):
 
 
 def uploadblobsfromrevs(repo, revs):
-    """upload lfs blobs introduced by revs
-
-    Note: also used by other extensions e. g. infinitepush. avoid renaming.
-    """
+    """upload lfs blobs introduced by revs"""
     if _canskipupload(repo):
         return
     pointers = extractpointers(repo, revs)
@@ -405,11 +413,11 @@ def prepush(pushop):
     return uploadblobsfromrevs(pushop.repo, pushop.outgoing.missing)
 
 
-@eh.wrapfunction(exchange, b'push')
+@eh.wrapfunction(exchange, 'push')
 def push(orig, repo, remote, *args, **kwargs):
     """bail on push if the extension isn't enabled on remote when needed, and
     update the remote store based on the destination path."""
-    if b'lfs' in repo.requirements:
+    if req_mod.LFS_REQUIREMENT in repo.requirements:
         # If the remote peer is for a local repo, the requirement tests in the
         # base class method enforce lfs support.  Otherwise, some revisions in
         # this repo use lfs, and the remote repo needs the extension loaded.
@@ -435,7 +443,7 @@ def push(orig, repo, remote, *args, **kwargs):
 
 
 # when writing a bundle via "hg bundle" command, upload related LFS blobs
-@eh.wrapfunction(bundle2, b'writenewbundle')
+@eh.wrapfunction(bundle2, 'writenewbundle')
 def writenewbundle(
     orig, ui, repo, source, filename, bundletype, outgoing, *args, **kwargs
 ):
@@ -524,14 +532,20 @@ def uploadblobs(repo, pointers):
     remoteblob.writebatch(pointers, repo.svfs.lfslocalblobstore)
 
 
-@eh.wrapfunction(upgrade_engine, b'finishdatamigration')
-def upgradefinishdatamigration(orig, ui, srcrepo, dstrepo, requirements):
+@eh.wrapfunction(upgrade_engine, 'finishdatamigration')
+def upgradefinishdatamigration(
+    orig: Callable[[UiT, RepoT, RepoT, RequirementSetT], None],
+    ui: UiT,
+    srcrepo: RepoT,
+    dstrepo: RepoT,
+    requirements: RequirementSetT,
+) -> None:
     orig(ui, srcrepo, dstrepo, requirements)
 
     # Skip if this hasn't been passed to reposetup()
-    if util.safehasattr(
-        srcrepo.svfs, b'lfslocalblobstore'
-    ) and util.safehasattr(dstrepo.svfs, b'lfslocalblobstore'):
+    if hasattr(srcrepo.svfs, 'lfslocalblobstore') and hasattr(
+        dstrepo.svfs, 'lfslocalblobstore'
+    ):
         srclfsvfs = srcrepo.svfs.lfslocalblobstore.vfs
         dstlfsvfs = dstrepo.svfs.lfslocalblobstore.vfs
 
@@ -541,10 +555,10 @@ def upgradefinishdatamigration(orig, ui, srcrepo, dstrepo, requirements):
                 lfutil.link(srclfsvfs.join(oid), dstlfsvfs.join(oid))
 
 
-@eh.wrapfunction(upgrade_actions, b'preservedrequirements')
-@eh.wrapfunction(upgrade_actions, b'supporteddestrequirements')
+@eh.wrapfunction(upgrade_actions, 'preservedrequirements')
+@eh.wrapfunction(upgrade_actions, 'supporteddestrequirements')
 def upgraderequirements(orig, repo):
     reqs = orig(repo)
-    if b'lfs' in repo.requirements:
-        reqs.add(b'lfs')
+    if req_mod.LFS_REQUIREMENT in repo.requirements:
+        reqs.add(req_mod.LFS_REQUIREMENT)
     return reqs
