@@ -11,22 +11,26 @@ This provides a read-only repository interface to bundles as if they
 were part of the actual repository.
 """
 
+from __future__ import annotations
 
 import contextlib
 import os
 import shutil
+import typing
 
 from .i18n import _
 from .node import (
     hex,
     nullrev,
 )
+from .interfaces.types import (
+    RevnumT,
+)
 
 from . import (
     bundle2,
     changegroup,
     changelog,
-    cmdutil,
     discovery,
     encoding,
     error,
@@ -38,6 +42,7 @@ from . import (
     pathutil,
     phases,
     pycompat,
+    repo as repo_utils,
     revlog,
     revlogutils,
     util,
@@ -51,9 +56,27 @@ from .revlogutils import (
     constants as revlog_constants,
 )
 
+from .interfaces import (
+    changegroup as i_cg,
+    compression as i_comp,
+)
+
 
 class bundlerevlog(revlog.revlog):
-    def __init__(self, opener, target, radix, cgunpacker, linkmapper):
+    # XXX this work around the buggy baserev in bundle revlog,
+    # We should fix that instead
+    may_emit_compressed = False
+
+    def __init__(
+        self,
+        opener: typing.Any,
+        target,
+        radix,
+        cgunpacker: i_cg.IChangeGroupUnpacker,
+        linkmapper,
+    ):
+        # TODO: figure out real type of opener
+        #
         # How it works:
         # To retrieve a revision, we need to know the offset of the revision in
         # the bundle (an unbundle object). We store this offset in the index
@@ -62,48 +85,56 @@ class bundlerevlog(revlog.revlog):
         # To differentiate a rev in the bundle from a rev in the revlog, we
         # check revision against repotiprev.
         opener = vfsmod.readonlyvfs(opener)
-        revlog.revlog.__init__(self, opener, target=target, radix=radix)
+        revlog.revlog.__init__(
+            self,
+            opener,
+            target=target,
+            radix=radix,
+            writable=False,
+        )
         self.bundle = cgunpacker
         n = len(self)
         self.repotiprev = n - 1
         self.bundlerevs = set()  # used by 'bundle()' revset expression
         for deltadata in cgunpacker.deltaiter():
-            node, p1, p2, cs, deltabase, delta, flags, sidedata = deltadata
-
-            size = len(delta)
+            baserev = self.rev(deltadata.delta_base)
+            base_size = self.rawsize(baserev)
+            full_size = mdiff.patchedsize(base_size, deltadata.delta)
+            size = len(deltadata.delta)
             start = cgunpacker.tell() - size
 
-            if self.index.has_node(node):
+            if self.index.has_node(deltadata.node):
                 # this can happen if two branches make the same change
-                self.bundlerevs.add(self.index.rev(node))
+                self.bundlerevs.add(self.index.rev(deltadata.node))
                 continue
-            if cs == node:
+            if deltadata.link_node == deltadata.node:
                 linkrev = nullrev
             else:
-                linkrev = linkmapper(cs)
+                linkrev = linkmapper(deltadata.link_node)
 
-            for p in (p1, p2):
+            for p in (deltadata.p1, deltadata.p2):
                 if not self.index.has_node(p):
                     raise error.LookupError(
                         p, self.display_id, _(b"unknown parent")
                     )
 
-            if not self.index.has_node(deltabase):
+            if not self.index.has_node(deltadata.delta_base):
                 raise error.LookupError(
-                    deltabase, self.display_id, _(b'unknown delta base')
+                    deltadata.delta_base,
+                    self.display_id,
+                    _(b'unknown delta base'),
                 )
 
-            baserev = self.rev(deltabase)
-            # start, size, full unc. size, base (unused), link, p1, p2, node, sidedata_offset (unused), sidedata_size (unused)
             e = revlogutils.entry(
-                flags=flags,
+                flags=deltadata.flags,
                 data_offset=start,
                 data_compressed_length=size,
+                data_uncompressed_length=full_size,
                 data_delta_base=baserev,
                 link_rev=linkrev,
-                parent_rev_1=self.rev(p1),
-                parent_rev_2=self.rev(p2),
-                node_id=node,
+                parent_rev_1=self.rev(deltadata.p1),
+                parent_rev_2=self.rev(deltadata.p2),
+                node_id=deltadata.node,
             )
             self.index.append(e)
             self.bundlerevs.add(n)
@@ -117,16 +148,27 @@ class bundlerevlog(revlog.revlog):
             with super().reading() as x:
                 yield x
 
+    def raw_comp_chunk(self, rev) -> tuple[i_comp.RevlogCompHeader, bytes]:
+        if rev <= self.repotiprev:
+            return super().raw_comp_chunk(rev)
+        self.bundle.seek(self.start(rev))
+        return (i_comp.REVLOG_COMP_NONE, self.bundle.read(self.length(rev)))
+
     def _chunk(self, rev):
         # Warning: in case of bundle, the diff is against what we stored as
         # delta base, not against rev - 1
         # XXX: could use some caching
         if rev <= self.repotiprev:
-            return revlog.revlog._chunk(self, rev)
+            return super()._inner._chunk(rev)
         self.bundle.seek(self.start(rev))
         return self.bundle.read(self.length(rev))
 
-    def revdiff(self, rev1, rev2):
+    def revdiff(
+        self,
+        rev1: RevnumT,
+        rev2: RevnumT,
+        extra_delta: bytes | None = None,
+    ):
         """return or calculate a delta between two revisions"""
         if rev1 > self.repotiprev and rev2 > self.repotiprev:
             # hot path for bundle
@@ -134,9 +176,19 @@ class bundlerevlog(revlog.revlog):
             if revb == rev1:
                 return self._chunk(rev2)
         elif rev1 <= self.repotiprev and rev2 <= self.repotiprev:
-            return revlog.revlog.revdiff(self, rev1, rev2)
+            return revlog.revlog.revdiff(self, rev1, rev2, extra_delta)
 
-        return mdiff.textdiff(self.rawdata(rev1), self.rawdata(rev2))
+        old = self.rawdata(rev1, validate=False)
+        new = self.rawdata(rev2, validate=False)
+        if extra_delta is not None:
+            base = new
+            new = mdiff.full_text_from_delta(
+                extra_delta,
+                len(new),
+                lambda: base,
+            )
+
+        return self._diff_fn(old, new)
 
     def _rawtext(self, node, rev):
         if rev is None:
@@ -147,18 +199,16 @@ class bundlerevlog(revlog.revlog):
         iterrev = rev
         # reconstruct the revision if it is from a changegroup
         while iterrev > self.repotiprev:
-            if (
-                self._inner._revisioncache
-                and self._inner._revisioncache[1] == iterrev
-            ):
-                rawtext = self._inner._revisioncache[2]
+            cached = self._inner.get_cached_text(iterrev)
+            if cached is not None:
+                rawtext = cached[1]
                 break
             chain.append(iterrev)
             iterrev = self.index[iterrev][3]
         if iterrev == nullrev:
             rawtext = b''
         elif rawtext is None:
-            r = super(bundlerevlog, self)._rawtext(
+            r = super()._rawtext(
                 self.node(iterrev),
                 iterrev,
             )
@@ -236,12 +286,12 @@ class bundlemanifest(bundlerevlog, manifest.manifestrevlog):
                 self._dirlogstarts,
                 dir=d,
             )
-        return super(bundlemanifest, self).dirlog(d)
+        return super().dirlog(d)
 
 
 class bundlefilelog(filelog.filelog):
-    def __init__(self, opener, path, cgunpacker, linkmapper):
-        filelog.filelog.__init__(self, opener, path)
+    def __init__(self, opener, path, radix, cgunpacker, linkmapper):
+        filelog.filelog.__init__(self, opener, path, radix, writable=False)
         self._revlog = bundlerevlog(
             opener,
             # XXX should use the unencoded path
@@ -259,14 +309,14 @@ class bundlepeer(localrepo.localpeer):
 
 class bundlephasecache(phases.phasecache):
     def __init__(self, *args, **kwargs):
-        super(bundlephasecache, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         if hasattr(self, 'opener'):
             self.opener = vfsmod.readonlyvfs(self.opener)
 
-    def write(self):
+    def write(self, repo):
         raise NotImplementedError
 
-    def _write(self, fp):
+    def _write(self, repo, fp):
         raise NotImplementedError
 
     def _updateroots(self, repo, phase, newroots, tr, invalidate=True):
@@ -276,17 +326,23 @@ class bundlephasecache(phases.phasecache):
         self.dirty = True
 
 
-def _getfilestarts(cgunpacker):
+def _getfilestarts(cgunpacker: i_cg.IChangeGroupUnpacker):
     filespos = {}
     for chunkdata in iter(cgunpacker.filelogheader, {}):
         fname = chunkdata[b'filename']
         filespos[fname] = cgunpacker.tell()
-        for chunk in iter(lambda: cgunpacker.deltachunk(None), {}):
+        for chunk in iter(lambda: cgunpacker.deltachunk(None), None):
             pass
     return filespos
 
 
-class bundlerepository:
+_bundle_repo_baseclass = object
+
+if typing.TYPE_CHECKING:
+    _bundle_repo_baseclass = localrepo.localrepository
+
+
+class bundlerepository(_bundle_repo_baseclass):
     """A repository instance that is a union of a local repo and a bundle.
 
     Instances represent a read-only repository composed of a local repository
@@ -299,6 +355,8 @@ class bundlerepository:
     Use instance() or makebundlerepository() to create instances.
     """
 
+    is_bundle_repo = True
+
     def __init__(self, bundlepath, url, tempparent):
         self._tempparent = tempparent
         self._url = url
@@ -308,7 +366,7 @@ class bundlerepository:
         # dict with the mapping 'filename' -> position in the changegroup.
         self._cgfilespos = {}
         self._bundlefile = None
-        self._cgunpacker = None
+        self._cgunpacker: i_cg.IChangeGroupUnpacker = None
         self.tempfile = None
         f = util.posixfile(bundlepath, b"rb")
         bundle = exchange.readbundle(self.ui, f, bundlepath)
@@ -325,7 +383,7 @@ class bundlerepository:
                         raise NotImplementedError(
                             b"can't process multiple changegroups"
                         )
-                    cgpart = part
+                    cgpart: bundle2.seekableunbundlepart = part.as_seekable()
                     self._handle_bundle2_cg_part(bundle, part)
 
             if not cgpart:
@@ -345,10 +403,15 @@ class bundlerepository:
                 _(b'bundle type %r cannot be read') % type(bundle)
             )
 
-    def _handle_bundle1(self, bundle, bundlepath):
+    def _handle_bundle1(self, bundle: changegroup.cg1unpacker, bundlepath):
         if bundle.compressed():
             f = self._writetempbundle(bundle.read, b'.hg10un', header=b'HG10UN')
-            bundle = exchange.readbundle(self.ui, f, bundlepath, self.vfs)
+            bundle: changegroup.cg1unpacker = typing.cast(
+                # XXX: We know what this is as we just wrote it,
+                # using a cast is lazy, but simple. Fix me later.
+                changegroup.cg1unpacker,
+                exchange.readbundle(self.ui, f, bundlepath, self.vfs),
+            )
 
         self._bundlefile = bundle
         self._cgunpacker = bundle
@@ -408,7 +471,7 @@ class bundlerepository:
         with os.fdopen(fdtemp, 'wb') as fptemp:
             fptemp.write(header)
             while True:
-                chunk = readfn(2 ** 18)
+                chunk = readfn(2**18)
                 if not chunk:
                     break
                 fptemp.write(chunk)
@@ -477,7 +540,9 @@ class bundlerepository:
     def url(self):
         return self._url
 
-    def file(self, f):
+    def file(self, f, writable=False):
+        assert not writable, "bundlerepository is read-only"
+
         if not self._cgfilespos:
             self._cgunpacker.seek(self.filestart)
             self._cgfilespos = _getfilestarts(self._cgunpacker)
@@ -485,9 +550,16 @@ class bundlerepository:
         if f in self._cgfilespos:
             self._cgunpacker.seek(self._cgfilespos[f])
             linkmapper = self.unfiltered().changelog.rev
-            return bundlefilelog(self.svfs, f, self._cgunpacker, linkmapper)
+            radix = self.store.filelog_radix_for_reading(f)
+            return bundlefilelog(
+                self.svfs,
+                f,
+                radix,
+                self._cgunpacker,
+                linkmapper,
+            )
         else:
-            return super(bundlerepository, self).file(f)
+            return super().file(f, writable=writable)
 
     def close(self):
         """Close assigned bundle file immediately."""
@@ -517,7 +589,7 @@ class bundlerepository:
             self.ui.warn(msg % hex(p1))
         if self.changelog.repotiprev < p2rev:
             self.ui.warn(msg % hex(p2))
-        return super(bundlerepository, self).setparents(p1, p2)
+        return super().setparents(p1, p2)
 
 
 def instance(ui, path, create, intents=None, createopts=None):
@@ -527,7 +599,7 @@ def instance(ui, path, create, intents=None, createopts=None):
     parentpath = ui.config(b"bundle", b"mainreporoot")
     if not parentpath:
         # try to find the correct path to the working directory repo
-        parentpath = cmdutil.findrepo(encoding.getcwd())
+        parentpath = repo_utils.find_repo(encoding.getcwd())
         if parentpath is None:
             parentpath = b''
     if parentpath:
@@ -605,6 +677,36 @@ class bundletransactionmanager:
         raise NotImplementedError
 
 
+class getremotechanges_state_tracker:
+    def __init__(self, peer, incoming, common, rheads):
+        # bundle file to be deleted
+        self.bundle = None
+        # bundle repo to be closed
+        self.bundlerepo = None
+        # remote peer connection to be closed
+        self.peer = peer
+        # if peer is remote, `localrepo` will be equal to
+        # `bundlerepo` when bundle is created.
+        self.localrepo = peer.local()
+
+        # `incoming` operation parameters:
+        # (these get mutated by _create_bundle)
+        self.incoming = incoming
+        self.common = common
+        self.rheads = rheads
+
+    def cleanup(self):
+        try:
+            if self.bundlerepo:
+                self.bundlerepo.close()
+        finally:
+            try:
+                if self.bundle:
+                    os.unlink(self.bundle)
+            finally:
+                self.peer.close()
+
+
 def getremotechanges(
     ui, repo, peer, onlyheads=None, bundlename=None, force=False
 ):
@@ -640,101 +742,127 @@ def getremotechanges(
     commonset = set(common)
     rheads = [x for x in rheads if x not in commonset]
 
-    bundle = None
-    bundlerepo = None
-    localrepo = peer.local()
-    if bundlename or not localrepo:
-        # create a bundle (uncompressed if peer repo is not local)
+    state = getremotechanges_state_tracker(peer, incoming, common, rheads)
 
-        # developer config: devel.legacy.exchange
-        legexc = ui.configlist(b'devel', b'legacy.exchange')
-        forcebundle1 = b'bundle2' not in legexc and b'bundle1' in legexc
-        canbundle2 = (
-            not forcebundle1
-            and peer.capable(b'getbundle')
-            and peer.capable(b'bundle2')
+    try:
+        csets = _getremotechanges_slowpath(
+            state, ui, repo, bundlename=bundlename, onlyheads=onlyheads
         )
-        if canbundle2:
-            with peer.commandexecutor() as e:
-                b2 = e.callcommand(
+        return (state.localrepo, csets, state.cleanup)
+    except:  # re-raises
+        state.cleanup()
+        raise
+
+
+def _create_bundle(state, ui, repo, bundlename, onlyheads):
+    # create a bundle (uncompressed if peer repo is not local)
+
+    # developer config: devel.legacy.exchange
+    legexc = ui.configlist(b'devel', b'legacy.exchange')
+    forcebundle1 = b'bundle2' not in legexc and b'bundle1' in legexc
+    canbundle2 = (
+        not forcebundle1
+        and state.peer.capable(b'getbundle')
+        and state.peer.capable(b'bundle2')
+    )
+    if canbundle2:
+        with state.peer.commandexecutor() as e:
+            b2 = e.callcommand(
+                b'getbundle',
+                {
+                    b'source': b'incoming',
+                    b'common': state.common,
+                    b'heads': state.rheads,
+                    b'bundlecaps': exchange.caps20to10(repo, role=b'client'),
+                    b'cg': True,
+                },
+            ).result()
+
+            fname = state.bundle = changegroup.writechunks(
+                ui, b2._forwardchunks(), bundlename
+            )
+    else:
+        if state.peer.capable(b'getbundle'):
+            with state.peer.commandexecutor() as e:
+                cg = e.callcommand(
                     b'getbundle',
                     {
                         b'source': b'incoming',
-                        b'common': common,
-                        b'heads': rheads,
-                        b'bundlecaps': exchange.caps20to10(
-                            repo, role=b'client'
-                        ),
-                        b'cg': True,
+                        b'common': state.common,
+                        b'heads': state.rheads,
+                    },
+                ).result()
+        elif onlyheads is None and not state.peer.capable(b'changegroupsubset'):
+            # compat with older servers when pulling all remote heads
+
+            with state.peer.commandexecutor() as e:
+                cg = e.callcommand(
+                    b'changegroup',
+                    {
+                        b'nodes': state.incoming,
+                        b'source': b'incoming',
                     },
                 ).result()
 
-                fname = bundle = changegroup.writechunks(
-                    ui, b2._forwardchunks(), bundlename
-                )
+            state.rheads = None
         else:
-            if peer.capable(b'getbundle'):
-                with peer.commandexecutor() as e:
-                    cg = e.callcommand(
-                        b'getbundle',
-                        {
-                            b'source': b'incoming',
-                            b'common': common,
-                            b'heads': rheads,
-                        },
-                    ).result()
-            elif onlyheads is None and not peer.capable(b'changegroupsubset'):
-                # compat with older servers when pulling all remote heads
+            with state.peer.commandexecutor() as e:
+                cg = e.callcommand(
+                    b'changegroupsubset',
+                    {
+                        b'bases': state.incoming,
+                        b'heads': state.rheads,
+                        b'source': b'incoming',
+                    },
+                ).result()
 
-                with peer.commandexecutor() as e:
-                    cg = e.callcommand(
-                        b'changegroup',
-                        {
-                            b'nodes': incoming,
-                            b'source': b'incoming',
-                        },
-                    ).result()
+        if state.localrepo:
+            bundletype = b"HG10BZ"
+        else:
+            bundletype = b"HG10UN"
+        fname = state.bundle = bundle2.writebundle(
+            ui, cg, bundlename, bundletype
+        )
+    # keep written bundle?
+    if bundlename:
+        state.bundle = None
 
-                rheads = None
-            else:
-                with peer.commandexecutor() as e:
-                    cg = e.callcommand(
-                        b'changegroupsubset',
-                        {
-                            b'bases': incoming,
-                            b'heads': rheads,
-                            b'source': b'incoming',
-                        },
-                    ).result()
+    return fname
 
-            if localrepo:
-                bundletype = b"HG10BZ"
-            else:
-                bundletype = b"HG10UN"
-            fname = bundle = bundle2.writebundle(ui, cg, bundlename, bundletype)
-        # keep written bundle?
-        if bundlename:
-            bundle = None
-        if not localrepo:
+
+def _getremotechanges_slowpath(
+    state, ui, repo, bundlename=None, onlyheads=None
+):
+    if bundlename or not state.localrepo:
+        fname = _create_bundle(
+            state,
+            ui,
+            repo,
+            bundlename=bundlename,
+            onlyheads=onlyheads,
+        )
+        if not state.localrepo:
             # use the created uncompressed bundlerepo
-            localrepo = bundlerepo = makebundlerepository(
+            state.localrepo = state.bundlerepo = makebundlerepository(
                 repo.baseui, repo.root, fname
             )
 
             # this repo contains local and peer now, so filter out local again
-            common = repo.heads()
-    if localrepo:
+            state.common = repo.heads()
+
+    if state.localrepo:
         # Part of common may be remotely filtered
         # So use an unfiltered version
         # The discovery process probably need cleanup to avoid that
-        localrepo = localrepo.unfiltered()
+        state.localrepo = state.localrepo.unfiltered()
 
-    csets = localrepo.changelog.findmissing(common, rheads)
+    csets = state.localrepo.changelog.findmissing(state.common, state.rheads)
 
-    if bundlerepo:
+    if state.bundlerepo:
+        bundlerepo = state.bundlerepo
         reponodes = [ctx.node() for ctx in bundlerepo[bundlerepo.firstnewrev :]]
 
-        with peer.commandexecutor() as e:
+        with state.peer.commandexecutor() as e:
             remotephases = e.callcommand(
                 b'listkeys',
                 {
@@ -743,16 +871,9 @@ def getremotechanges(
             ).result()
 
         pullop = exchange.pulloperation(
-            bundlerepo, peer, path=None, heads=reponodes
+            bundlerepo, state.peer, path=None, heads=reponodes
         )
         pullop.trmanager = bundletransactionmanager()
         exchange._pullapplyphases(pullop, remotephases)
 
-    def cleanup():
-        if bundlerepo:
-            bundlerepo.close()
-        if bundle:
-            os.unlink(bundle)
-        peer.close()
-
-    return (localrepo, csets, cleanup)
+    return csets
