@@ -8,17 +8,18 @@
 # This code is based on the Mark Edgington's crecord extension.
 # (Itself based on Bryan O'Sullivan's record extension.)
 
-from __future__ import absolute_import
+from __future__ import annotations
 
 import os
 import re
 import signal
 
-from .i18n import _
-from .pycompat import (
-    getattr,
-    open,
+from typing import (
+    Any,
+    Optional,
 )
+
+from .i18n import _
 from . import (
     diffhelper,
     encoding,
@@ -34,14 +35,14 @@ stringio = util.stringio
 
 # patch comments based on the git one
 diffhelptext = _(
-    """# To remove '-' lines, make them ' ' lines (context).
+    b"""# To remove '-' lines, make them ' ' lines (context).
 # To remove '+' lines, delete them.
 # Lines starting with # will be removed from the patch.
 """
 )
 
 hunkhelptext = _(
-    """#
+    b"""#
 # If the patch applies cleanly, the edited hunk will immediately be
 # added to the record list. If it does not apply cleanly, a rejects file
 # will be generated. You can use that when you try again. If all lines
@@ -51,7 +52,7 @@ hunkhelptext = _(
 )
 
 patchhelptext = _(
-    """#
+    b"""#
 # If the patch applies cleanly, the edited patch will immediately
 # be finalised. If it does not apply cleanly, rejects files will be
 # generated. You can use those when you try again.
@@ -64,7 +65,7 @@ try:
 
     curses.error
 except (ImportError, AttributeError):
-    curses = False
+    curses = None
 
 
 class fallbackerror(error.Abort):
@@ -83,11 +84,17 @@ def checkcurses(ui):
     return curses and ui.interface(b"chunkselector") == b"curses"
 
 
-class patchnode(object):
+class patchnode:
     """abstract class for patch graph nodes
     (i.e. patchroot, header, hunk, hunkline)
     """
 
+    folded: bool = False
+
+    @property
+    def content(self):
+        return b''
+
     def firstchild(self):
         raise NotImplementedError(b"method must be implemented by subclass")
 
@@ -128,11 +135,7 @@ class patchnode(object):
 
         If it is not possible to get the next item, return None.
         """
-        try:
-            itemfolded = self.folded
-        except AttributeError:
-            itemfolded = False
-        if skipfolded and itemfolded:
+        if skipfolded and self.folded:
             nextitem = self.nextsibling()
             if nextitem is None:
                 try:
@@ -163,7 +166,7 @@ class patchnode(object):
             except AttributeError:  # parent and/or grandparent was None
                 return None
 
-    def previtem(self):
+    def previtem(self, skipfolded=None):
         """
         Try to return the previous item closest to this item, regardless of
         item's type (header, hunk, or hunkline).
@@ -203,9 +206,8 @@ class patch(patchnode, list):  # todo: rename patchroot
 
 
 class uiheader(patchnode):
-    """patch header
-
-    xxx shouldn't we move this to mercurial/patch.py ?
+    """
+    patchnode class wrapping a patch.header
     """
 
     def __init__(self, header):
@@ -226,6 +228,10 @@ class uiheader(patchnode):
         self.neverunfolded = True
         self.hunks = [uihunk(h, self) for h in self.hunks]
 
+    @property
+    def content(self):
+        return self.filename()
+
     def prettystr(self):
         x = stringio()
         self.pretty(x)
@@ -292,6 +298,10 @@ class uihunkline(patchnode):
         # in the previtem method.
         self.folded = False
 
+    @property
+    def content(self):
+        return self.linetext
+
     def prettystr(self):
         return self.linetext
 
@@ -331,7 +341,7 @@ class uihunkline(patchnode):
 
 
 class uihunk(patchnode):
-    """ui patch hunk, wraps a hunk and keep track of ui behavior """
+    """ui patch hunk, wraps a hunk and keep track of ui behavior"""
 
     maxcontext = 3
 
@@ -350,6 +360,10 @@ class uihunk(patchnode):
         # children are partially applied (i.e. some applied, some not).
         self.partial = False
 
+    @property
+    def content(self):
+        return self.proc if self.proc else b''
+
     def nextsibling(self):
         numhunksinheader = len(self.header.hunks)
         indexofthishunk = self.header.hunks.index(self)
@@ -506,7 +520,7 @@ class uihunk(patchnode):
             text = line.linetext
             if line.linetext == diffhelper.MISSING_NEWLINE_MARKER:
                 noeol = True
-                break
+                continue
             if line.applied:
                 if text.startswith(b'+'):
                     dels.append(text[1:])
@@ -574,7 +588,7 @@ def chunkselector(ui, headerlist, operation=None):
     ui.write(_(b'starting interactive selection\n'))
     chunkselector = curseschunkselector(headerlist, ui, operation)
     origsigtstp = sentinel = object()
-    if util.safehasattr(signal, b'SIGTSTP'):
+    if hasattr(signal, 'SIGTSTP'):
         origsigtstp = signal.getsignal(signal.SIGTSTP)
     try:
         with util.with_lc_ctype():
@@ -602,7 +616,7 @@ def testchunkselector(testfn, ui, headerlist, operation=None):
     """
     chunkselector = curseschunkselector(headerlist, ui, operation)
 
-    class dummystdscr(object):
+    class dummystdscr:
         def clear(self):
             pass
 
@@ -611,9 +625,10 @@ def testchunkselector(testfn, ui, headerlist, operation=None):
 
     chunkselector.stdscr = dummystdscr()
     if testfn and os.path.exists(testfn):
-        testf = open(testfn, 'r')
-        testcommands = [x.rstrip('\n') for x in testf.readlines()]
-        testf.close()
+        # This file is purposely NOT opened in binary mode
+        with open(testfn) as testf:
+            testcommands = [x.rstrip('\n') for x in testf.readlines()]
+
         while True:
             if chunkselector.handlekeypressed(testcommands.pop(0), test=True):
                 break
@@ -628,7 +643,16 @@ _headermessages = {  # {operation: text}
 }
 
 
-class curseschunkselector(object):
+class curseschunkselector:
+    # Attributes not set by __init__ but by `main`
+    yscreensize: int
+    xscreensize: int
+    initexc: Optional[Exception]
+    # These are actually a `curses.window` object, but since curses is
+    # Optional we can't use it.
+    chunkpad: Any
+    statuswin: Any
+
     def __init__(self, headerlist, ui, operation=None):
         # put the headers into a patch object
         self.headerlist = patch(headerlist)
@@ -695,6 +719,8 @@ class curseschunkselector(object):
             )
         self.operation = operation
 
+        self.regex = None
+
     def uparrowevent(self):
         """
         try to select the previous item to the current item that has the
@@ -1151,7 +1177,7 @@ class curseschunkselector(object):
             numtrailingspaces = origlen - strippedlen
 
         if towin:
-            window.addstr(text, colorpair)
+            window.addstr(encoding.strfromlocal(text), colorpair)
         t += text
 
         if showwhtspc:
@@ -1289,7 +1315,7 @@ class curseschunkselector(object):
         return checkbox
 
     def printheader(
-        self, header, selected=False, towin=True, ignorefolding=False
+        self, header, selected: bool = False, towin=True, ignorefolding=False
     ):
         """
         print the header to the pad.  if countlines is True, don't print
@@ -1307,7 +1333,8 @@ class curseschunkselector(object):
             )
         # select color-pair based on if the header is selected
         colorpair = self.getcolorpair(
-            name=selected and b"selected" or b"normal", attrlist=[curses.A_BOLD]
+            name=b"selected" if selected else b"normal",
+            attrlist=[curses.A_BOLD],
         )
 
         # print out each line of the chunk, expanding it to screen width
@@ -1334,7 +1361,7 @@ class curseschunkselector(object):
         return outstr
 
     def printhunklinesbefore(
-        self, hunk, selected=False, towin=True, ignorefolding=False
+        self, hunk, selected: bool = False, towin=True, ignorefolding=False
     ):
         """includes start/end line indicator"""
         outstr = b""
@@ -1348,7 +1375,8 @@ class curseschunkselector(object):
             )
 
         colorpair = self.getcolorpair(
-            name=selected and b"selected" or b"normal", attrlist=[curses.A_BOLD]
+            name=b"selected" if selected else b"normal",
+            attrlist=[curses.A_BOLD],
         )
 
         # print out from-to line with checkbox
@@ -1547,7 +1575,6 @@ class curseschunkselector(object):
     def sigwinchhandler(self, n, frame):
         """handle window resizing"""
         try:
-            curses.endwin()
             self.xscreensize, self.yscreensize = scmutil.termsize(self.ui)
             self.statuswin.resize(self.numstatuslines, self.xscreensize)
             self.numpadlines = self.getnumlinesdisplayed(ignorefolding=True) + 1
@@ -1621,7 +1648,7 @@ class curseschunkselector(object):
     def helpwindow(self):
         """print a help window to the screen.  exit after any keypress."""
         helptext = _(
-            """            [press any key to return to the patch-display]
+            b"""            [press any key to return to the patch-display]
 
 The curses hunk selector allows you to interactively choose among the
 changes you have made, and confirm only those changes you select for
@@ -1649,6 +1676,9 @@ smaller changesets. the following are valid keystrokes:
                       a : toggle all selections
                       c : confirm selected changes
                       r : review/edit and confirm selected changes
+                      / : regex search for code or filename
+                      n : next search result for code or filename
+                      N : previous search result for code or filename
                       q : quit without confirming (no changes will be made)
                       ? : help (what you're currently reading)"""
         )
@@ -1675,7 +1705,6 @@ smaller changesets. the following are valid keystrokes:
 
         curses.raw()
         curses.def_prog_mode()
-        curses.endwin()
         self.commenttext = self.ui.edit(self.commenttext, self.ui.username())
         curses.cbreak()
         self.stdscr.refresh()
@@ -1745,7 +1774,7 @@ smaller changesets. the following are valid keystrokes:
         """ask for 'y' to be pressed to confirm selected. return True if
         confirmed."""
         confirmtext = _(
-            """If you answer yes to the following, your currently chosen patch chunks
+            b"""If you answer yes to the following, your currently chosen patch chunks
 will be loaded into an editor. To modify the patch, make the changes in your
 editor and save. To accept the current patch as-is, close the editor without
 saving.
@@ -1873,6 +1902,86 @@ are you sure you want to review/edit and confirm the selected changes [yn]?
                 return False
         return True
 
+    def errorpopup(self, msg, win):
+        self.printstring(
+            win,
+            msg,
+            pairname=b"legend",
+            align=False,
+        )
+        while win.getkey() not in ["\n", "KEY_ENTER"]:
+            pass
+
+    def handlesearch(self):
+        win = curses.newwin(1, self.xscreensize, self.yscreensize - 1, 0)
+        win.echochar("/")
+
+        curses.echo()
+        curses.curs_set(1)
+        self.regex = win.getstr() or None
+        curses.noecho()
+        curses.curs_set(0)
+
+        try:
+            if not self.showsearch(self.regex):
+                self.errorpopup(_(b"Pattern not found (press ENTER)"), win)
+        except re.error as err:
+            msg = _(b'Bad regex: %s (press ENTER)') % err.msg.encode()
+            self.errorpopup(msg, win)
+
+        del win
+
+        self.stdscr.clear()
+        self.stdscr.refresh()
+
+    def showsearch(self, regex, forward=True):
+        if not regex:
+            return
+
+        moveattr = 'nextitem' if forward else 'previtem'
+        currentitem = getattr(self.currentselecteditem, moveattr)(
+            skipfolded=False
+        )
+
+        matches = None
+        regex = re.compile(regex)
+        while currentitem:
+            matches = regex.search(currentitem.content)
+            if matches:
+                self.currentselecteditem = currentitem
+                break
+            currentitem = getattr(currentitem, moveattr)(skipfolded=False)
+
+        # Whatever is selected should now be visible
+        unfolded = self.currentselecteditem
+        while matches and unfolded:
+            unfolded.folded = False
+            unfolded = unfolded.parentitem()
+
+        return matches
+
+    def searchdirection(self, failuremsg, forward=True):
+        if not self.regex:
+            return
+
+        if not self.showsearch(self.regex, forward=forward):
+            win = curses.newwin(1, self.xscreensize, self.yscreensize - 1, 0)
+            self.errorpopup(failuremsg, win)
+            del win
+
+            self.stdscr.clear()
+            self.stdscr.refresh()
+
+    def handlenextsearch(self):
+        self.searchdirection(
+            _(b"Next pattern not found (press ENTER)"), forward=True
+        )
+
+    def handleprevsearch(self):
+        self.searchdirection(
+            _(b"Previous pattern not found (press ENTER)"), forward=False
+        )
+
     def handlekeypressed(self, keypressed, test=False):
         """
         Perform actions based on pressed keys.
@@ -1923,6 +2032,12 @@ are you sure you want to review/edit and confirm the selected changes [yn]?
             self.togglefolded(foldparent=True)
         elif keypressed in ["m"]:
             self.commitMessageWindow()
+        elif keypressed in ["/"]:
+            self.handlesearch()
+        elif keypressed in ["n"]:
+            self.handlenextsearch()
+        elif keypressed in ["N"]:
+            self.handleprevsearch()
         elif keypressed in ["g", "KEY_HOME"]:
             self.handlefirstlineevent()
         elif keypressed in ["G", "KEY_END"]:
@@ -1944,7 +2059,7 @@ are you sure you want to review/edit and confirm the selected changes [yn]?
         """
 
         origsigwinch = sentinel = object()
-        if util.safehasattr(signal, b'SIGWINCH'):
+        if hasattr(signal, 'SIGWINCH'):
             origsigwinch = signal.signal(signal.SIGWINCH, self.sigwinchhandler)
         try:
             return self._main(stdscr)
@@ -1990,7 +2105,7 @@ are you sure you want to review/edit and confirm the selected changes [yn]?
         )
         # newwin([height, width,] begin_y, begin_x)
         self.statuswin = curses.newwin(self.numstatuslines, 0, 0, 0)
-        self.statuswin.keypad(1)  # interpret arrow-key, etc. esc sequences
+        self.statuswin.keypad(True)  # interpret arrow-key, etc. esc sequences
 
         # figure out how much space to allocate for the chunk-pad which is
         # used for displaying the patch
