@@ -5,14 +5,26 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
+from __future__ import annotations
 
 import collections
 import contextlib
 import os
 import stat
+import typing
 import uuid
 
+from typing import (
+    Any,
+    Iterable,
+    Iterator,
+)
+
 from .i18n import _
+from .interfaces.types import (
+    MatcherT,
+    TransactionT,
+)
 
 from hgdemandimport import tracing
 
@@ -36,9 +48,23 @@ from .dirstateutils import (
 
 from .interfaces import (
     dirstate as intdirstate,
-    util as interfaceutil,
 )
 
+if typing.TYPE_CHECKING:
+    from typing import (
+        Callable,
+        TypeVar,
+    )
+
+    # For some reason, pytype got confused by using ParamSpec, and demoted
+    # some (but not all) decorated methods to Callable with no defined args.
+    # This manages to preserve all methods, even including the variable names.
+    _C = TypeVar('_C', bound=Callable)
+    """An unconstrained Callable, for when a method takes a method with any
+    signature, and returns a method with an identical signature.
+    """
+
+
 parsers = policy.importmod('parsers')
 rustmod = policy.importrust('dirstate')
 
@@ -65,7 +91,7 @@ class rootcache(filecache):
         return obj._join(fname)
 
 
-def check_invalidated(func):
+def check_invalidated(func: _C) -> _C:
     """check that the func is called with a non-invalidated dirstate
 
     The dirstate is in an "invalidated state" after an error occured during its
@@ -83,7 +109,7 @@ def check_invalidated(func):
     return wrap
 
 
-def requires_changing_parents(func):
+def requires_changing_parents(func: _C) -> _C:
     def wrap(self, *args, **kwargs):
         if not self.is_changing_parents:
             msg = 'calling `%s` outside of a changing_parents context'
@@ -94,7 +120,7 @@ def requires_changing_parents(func):
     return check_invalidated(wrap)
 
 
-def requires_changing_files(func):
+def requires_changing_files(func: _C) -> _C:
     def wrap(self, *args, **kwargs):
         if not self.is_changing_files:
             msg = 'calling `%s` outside of a `changing_files`'
@@ -105,7 +131,7 @@ def requires_changing_files(func):
     return check_invalidated(wrap)
 
 
-def requires_changing_any(func):
+def requires_changing_any(func: _C) -> _C:
     def wrap(self, *args, **kwargs):
         if not self.is_changing_any:
             msg = 'calling `%s` outside of a changing context'
@@ -116,7 +142,7 @@ def requires_changing_any(func):
     return check_invalidated(wrap)
 
 
-def requires_changing_files_or_status(func):
+def requires_changing_files_or_status(func: _C) -> _C:
     def wrap(self, *args, **kwargs):
         if not (self.is_changing_files or self._running_status > 0):
             msg = (
@@ -134,9 +160,7 @@ CHANGE_TYPE_PARENTS = "parents"
 CHANGE_TYPE_FILES = "files"
 
 
-@interfaceutil.implementer(intdirstate.idirstate)
-class dirstate:
-
+class dirstate(intdirstate.idirstate):
     # used by largefile to avoid overwritting transaction callback
     _tr_key_suffix = b''
 
@@ -398,7 +422,7 @@ class dirstate:
         raise error.ProgrammingError(msg)
 
     @property
-    def is_changing_any(self):
+    def is_changing_any(self) -> bool:
         """Returns true if the dirstate is in the middle of a set of changes.
 
         This returns True for any kind of change.
@@ -406,7 +430,7 @@ class dirstate:
         return self._changing_level > 0
 
     @property
-    def is_changing_parents(self):
+    def is_changing_parents(self) -> bool:
         """Returns true if the dirstate is in the middle of a set of changes
         that modify the dirstate parent.
         """
@@ -415,7 +439,7 @@ class dirstate:
         return self._change_type == CHANGE_TYPE_PARENTS
 
     @property
-    def is_changing_files(self):
+    def is_changing_files(self) -> bool:
         """Returns true if the dirstate is in the middle of a set of changes
         that modify the files tracked or their sources.
         """
@@ -471,11 +495,12 @@ class dirstate:
     def _pl(self):
         return self._map.parents()
 
-    def hasdir(self, d):
+    def hasdir(self, d: bytes) -> bool:
         return self._map.hastrackeddir(d)
 
     @rootcache(b'.hgignore')
-    def _ignore(self):
+    @util.rust_tracing_span("dirstate._ignore")
+    def _ignore(self) -> MatcherT:
         files = self._ignorefiles()
         if not files:
             return matchmod.never()
@@ -488,11 +513,11 @@ class dirstate:
         return self._ui.configbool(b'ui', b'slash') and pycompat.ossep != b'/'
 
     @propertycache
-    def _checklink(self):
+    def _checklink(self) -> bool:
         return util.checklink(self._root)
 
     @propertycache
-    def _checkexec(self):
+    def _checkexec(self) -> bool:
         return bool(util.checkexec(self._root))
 
     @propertycache
@@ -504,7 +529,9 @@ class dirstate:
         # it's safe because f is always a relative path
         return self._rootdir + f
 
-    def flagfunc(self, buildfallback):
+    def flagfunc(
+        self, buildfallback: intdirstate.FlagFuncFallbackT
+    ) -> intdirstate.FlagFuncReturnT:
         """build a callable that returns flags associated with a filename
 
         The information is extracted from three possible layers:
@@ -516,7 +543,7 @@ class dirstate:
         # small hack to cache the result of buildfallback()
         fallback_func = []
 
-        def get_flags(x):
+        def get_flags(x: bytes) -> bytes:
             entry = None
             fallback_value = None
             try:
@@ -567,7 +594,7 @@ class dirstate:
             return forcecwd
         return encoding.getcwd()
 
-    def getcwd(self):
+    def getcwd(self) -> bytes:
         """Return the path from which a canonical path is calculated.
 
         This path should be used to resolve file patterns or to convert
@@ -587,7 +614,7 @@ class dirstate:
             # we're outside the repo. return an absolute path.
             return cwd
 
-    def pathto(self, f, cwd=None):
+    def pathto(self, f: bytes, cwd: bytes | None = None) -> bytes:
         if cwd is None:
             cwd = self.getcwd()
         path = util.pathto(self._root, cwd, f)
@@ -595,31 +622,31 @@ class dirstate:
             return util.pconvert(path)
         return path
 
-    def get_entry(self, path):
+    def get_entry(self, path: bytes) -> intdirstate.DirstateItemT:
         """return a DirstateItem for the associated path"""
         entry = self._map.get(path)
         if entry is None:
             return DirstateItem()
         return entry
 
-    def __contains__(self, key):
+    def __contains__(self, key: Any) -> bool:
         return key in self._map
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[bytes]:
         return iter(sorted(self._map))
 
-    def items(self):
+    def items(self) -> Iterator[tuple[bytes, intdirstate.DirstateItemT]]:
         return self._map.items()
 
     iteritems = items
 
-    def parents(self):
+    def parents(self) -> list[bytes]:
         return [self._validate(p) for p in self._pl]
 
-    def p1(self):
+    def p1(self) -> bytes:
         return self._validate(self._pl[0])
 
-    def p2(self):
+    def p2(self) -> bytes:
         return self._validate(self._pl[1])
 
     @property
@@ -627,11 +654,11 @@ class dirstate:
         """True if a merge is in progress"""
         return self._pl[1] != self._nodeconstants.nullid
 
-    def branch(self):
+    def branch(self) -> bytes:
         return encoding.tolocal(self._branch)
 
     @requires_changing_parents
-    def setparents(self, p1, p2=None):
+    def setparents(self, p1: bytes, p2: bytes | None = None):
         """Set dirstate parents to p1 and p2.
 
         When moving from two parents to one, "merged" entries a
@@ -644,8 +671,8 @@ class dirstate:
             p2 = self._nodeconstants.nullid
         if self._changing_level == 0:
             raise ValueError(
-                b"cannot set dirstate parent outside of "
-                b"dirstate.changing_parents context manager"
+                "cannot set dirstate parent outside of "
+                "dirstate.changing_parents context manager"
             )
 
         self._dirty = True
@@ -657,7 +684,9 @@ class dirstate:
         fold_p2 = oldp2 != nullid and p2 == nullid
         return self._map.setparents(p1, p2, fold_p2=fold_p2)
 
-    def setbranch(self, branch, transaction):
+    def setbranch(
+        self, branch: bytes, transaction: TransactionT | None
+    ) -> None:
         self.__class__._branch.set(self, encoding.fromlocal(branch))
         if transaction is not None:
             self._setup_tr_abort(transaction)
@@ -685,7 +714,12 @@ class dirstate:
     def _write_branch(self, file_obj):
         file_obj.write(self._branch + b'\n')
 
-    def invalidate(self):
+    def invalidate_cwd(self) -> None:
+        """invalidate the cached read of current working copy"""
+        if '_cwd' in self.__dict__:
+            del self._cwd
+
+    def invalidate(self) -> None:
         """Causes the next access to reread the dirstate.
 
         This is different from localrepo.invalidatedirstate() because it always
@@ -705,7 +739,7 @@ class dirstate:
         self._origpl = None
 
     @requires_changing_any
-    def copy(self, source, dest):
+    def copy(self, source: bytes | None, dest: bytes) -> None:
         """Mark dest as a copy of source. Unmark dest if source is None."""
         if source == dest:
             return
@@ -716,10 +750,10 @@ class dirstate:
         else:
             self._map.copymap.pop(dest, None)
 
-    def copied(self, file):
+    def copied(self, file: bytes) -> bytes | None:
         return self._map.copymap.get(file, None)
 
-    def copies(self):
+    def copies(self) -> dict[bytes, bytes]:
         return self._map.copymap
 
     @requires_changing_files
@@ -760,7 +794,7 @@ class dirstate:
         return ret
 
     @requires_changing_files_or_status
-    def set_clean(self, filename, parentfiledata):
+    def set_clean(self, filename: bytes, parentfiledata) -> None:
         """record that the current state of the file on disk is known to be clean"""
         self._dirty = True
         if not self._map[filename].tracked:
@@ -769,7 +803,7 @@ class dirstate:
         self._map.set_clean(filename, mode, size, mtime)
 
     @requires_changing_files_or_status
-    def set_possibly_dirty(self, filename):
+    def set_possibly_dirty(self, filename: bytes) -> None:
         """record that the current state of the file on disk is unknown"""
         self._dirty = True
         self._map.set_possibly_dirty(filename)
@@ -777,9 +811,9 @@ class dirstate:
     @requires_changing_parents
     def update_file_p1(
         self,
-        filename,
+        filename: bytes,
         p1_tracked,
-    ):
+    ) -> None:
         """Set a file as tracked in the parent (or not)
 
         This is to be called when adjust the dirstate to a new parent after an history
@@ -789,7 +823,7 @@ class dirstate:
         a `with dirstate.changing_parents(repo):` context.
         """
         if self.in_merge:
-            msg = b'update_file_reference should not be called when merging'
+            msg = 'update_file_reference should not be called when merging'
             raise error.ProgrammingError(msg)
         entry = self._map.get(filename)
         if entry is None:
@@ -817,13 +851,13 @@ class dirstate:
     @requires_changing_parents
     def update_file(
         self,
-        filename,
+        filename: bytes,
         wc_tracked,
         p1_tracked,
-        p2_info=False,
-        possibly_dirty=False,
+        p2_info: bool = False,
+        possibly_dirty: bool = False,
         parentfiledata=None,
-    ):
+    ) -> None:
         """update the information about a file in the dirstate
 
         This is to be called when the direstates parent changes to keep track
@@ -880,7 +914,6 @@ class dirstate:
         possibly_dirty=False,
         parentfiledata=None,
     ):
-
         # note: I do not think we need to double check name clash here since we
         # are in a update/merge case that should already have taken care of
         # this. The test agrees
@@ -986,7 +1019,9 @@ class dirstate:
                 )
         return folded
 
-    def normalize(self, path, isknown=False, ignoremissing=False):
+    def normalize(
+        self, path: bytes, isknown: bool = False, ignoremissing: bool = False
+    ) -> bytes:
         """
         normalize the case of a pathname when on a casefolding filesystem
 
@@ -1012,12 +1047,17 @@ class dirstate:
     # - its semantic is unclear
     # - do we really needs it ?
     @requires_changing_parents
-    def clear(self):
+    def clear(self) -> None:
         self._map.clear()
         self._dirty = True
 
     @requires_changing_parents
-    def rebuild(self, parent, allfiles, changedfiles=None):
+    def rebuild(
+        self,
+        parent: bytes,
+        allfiles: Iterable[bytes],  # TODO: more than iterable? (uses len())
+        changedfiles: Iterable[bytes] | None = None,
+    ) -> None:
         matcher = self._sparsematcher
         if matcher is not None and not matcher.always():
             # should not add non-matching files
@@ -1083,7 +1123,7 @@ class dirstate:
             on_abort,
         )
 
-    def write(self, tr):
+    def write(self, tr: TransactionT | None) -> None:
         if not self._dirty:
             return
         # make sure we don't request a write of invalidated content
@@ -1092,7 +1132,6 @@ class dirstate:
 
         write_key = self._use_tracked_hint and self._dirty_tracked_set
         if tr:
-
             self._setup_tr_abort(tr)
             self._attached_to_a_transaction = True
 
@@ -1134,7 +1173,9 @@ class dirstate:
         self._opener.unlink(self._filename_th)
         self._use_tracked_hint = False
 
-    def addparentchangecallback(self, category, callback):
+    def addparentchangecallback(
+        self, category: bytes, callback: intdirstate.AddParentChangeCallbackT
+    ) -> None:
         """add a callback to be called when the wd parents are changed
 
         Callback will be called with the following arguments:
@@ -1169,7 +1210,7 @@ class dirstate:
                 return True
         return False
 
-    def _ignorefiles(self):
+    def _ignorefiles(self) -> list[bytes]:
         files = []
         if os.path.exists(self._join(b'.hgignore')):
             files.append(self._join(b'.hgignore'))
@@ -1180,7 +1221,7 @@ class dirstate:
                 files.append(os.path.join(self._rootdir, util.expandpath(path)))
         return files
 
-    def _ignorefileandline(self, f):
+    def _ignorefileandline(self, f: bytes) -> intdirstate.IgnoreFileAndLineT:
         files = collections.deque(self._ignorefiles())
         visited = set()
         while files:
@@ -1286,7 +1327,7 @@ class dirstate:
                     badfn(ff, badtype(kind))
                     if nf in dmap:
                         results[nf] = None
-            except (OSError) as inst:
+            except OSError as inst:
                 # nf not found on disk - it is dirstate only
                 if nf in dmap:  # does it exactly match a missing file?
                     results[nf] = None
@@ -1338,7 +1379,14 @@ class dirstate:
 
         return results, dirsfound, dirsnotfound
 
-    def walk(self, match, subrepos, unknown, ignored, full=True):
+    def walk(
+        self,
+        match: MatcherT,
+        subrepos: Any,
+        unknown: bool,
+        ignored: bool,
+        full: bool = True,
+    ) -> intdirstate.WalkReturnT:
         """
         Walk recursively through the directory tree, finding all files
         matched by match.
@@ -1444,10 +1492,10 @@ class dirstate:
                         # interested in comparing it to files currently in the
                         # dmap -- therefore normalizefile is enough
                         nf = normalizefile(
-                            nd and (nd + b"/" + f) or f, True, True
+                            (nd + b"/" + f) if nd else f, True, True
                         )
                     else:
-                        nf = nd and (nd + b"/" + f) or f
+                        nf = (nd + b"/" + f) if nd else f
                     if nf not in results:
                         if kind == dirkind:
                             if not ignore(nf):
@@ -1529,7 +1577,14 @@ class dirstate:
                     results[next(iv)] = st
         return results
 
-    def _rust_status(self, matcher, list_clean, list_ignored, list_unknown):
+    def _rust_status(
+        self,
+        matcher,
+        list_clean,
+        list_ignored,
+        list_unknown,
+        empty_dirs_keep_files,
+    ):
         if self._sparsematchfn is not None:
             em = matchmod.exact(matcher.files())
             sm = matchmod.unionmatcher([self._sparsematcher, em])
@@ -1556,7 +1611,7 @@ class dirstate:
             unknown,
             warnings,
             bad,
-            traversed,
+            empty_dirs,
             dirty,
         ) = rustmod.status(
             self._map._map,
@@ -1568,34 +1623,14 @@ class dirstate:
             bool(list_ignored),
             bool(list_unknown),
             bool(matcher.traversedir),
+            bool(empty_dirs_keep_files),
         )
 
         self._dirty |= dirty
 
-        if matcher.traversedir:
-            for dir in traversed:
-                matcher.traversedir(dir)
-
         if self._ui.warn:
-            for item in warnings:
-                if isinstance(item, tuple):
-                    file_path, syntax = item
-                    msg = _(b"%s: ignoring invalid syntax '%s'\n") % (
-                        file_path,
-                        syntax,
-                    )
-                    self._ui.warn(msg)
-                else:
-                    msg = _(b"skipping unreadable pattern file '%s': %s\n")
-                    self._ui.warn(
-                        msg
-                        % (
-                            pathutil.canonpath(
-                                self._rootdir, self._rootdir, item
-                            ),
-                            b"No such file or directory",
-                        )
-                    )
+            for warning in warnings:
+                self._ui.warn(warning)
 
         for fn, message in sorted(bad):
             matcher.bad(fn, encoding.strtolocal(message))
@@ -1608,10 +1643,33 @@ class dirstate:
             unknown=unknown,
             ignored=ignored,
             clean=clean,
+            empty_dirs=empty_dirs if matcher.traversedir else None,
         )
         return (lookup, status)
 
-    def status(self, match, subrepos, ignored, clean, unknown):
+    def use_rust_status(
+        self,
+        subrepos: bool,
+    ) -> bool:
+        if rustmod is None:
+            return False
+        elif self._checkcase:
+            # Case-insensitive filesystems are not handled yet
+            return False
+        elif subrepos:
+            return False
+
+        return True
+
+    def status(
+        self,
+        match: MatcherT,
+        subrepos: bool,
+        ignored: bool,
+        clean: bool,
+        unknown: bool,
+        empty_dirs_keep_files: bool = False,
+    ) -> intdirstate.StatusReturnT:
         """Determine the status of the working copy relative to the
         dirstate and return a pair of (unsure, status), where status is of type
         scmutil.status and:
@@ -1637,29 +1695,6 @@ class dirstate:
         dmap = self._map
         dmap.preload()
 
-        use_rust = True
-
-        allowed_matchers = (
-            matchmod.alwaysmatcher,
-            matchmod.differencematcher,
-            matchmod.exactmatcher,
-            matchmod.includematcher,
-            matchmod.intersectionmatcher,
-            matchmod.nevermatcher,
-            matchmod.unionmatcher,
-        )
-
-        if rustmod is None:
-            use_rust = False
-        elif self._checkcase:
-            # Case-insensitive filesystems are not handled yet
-            use_rust = False
-        elif subrepos:
-            use_rust = False
-        elif not isinstance(match, allowed_matchers):
-            # Some matchers have yet to be implemented
-            use_rust = False
-
         # Get the time from the filesystem so we can disambiguate files that
         # appear modified in the present or future.
         try:
@@ -1668,10 +1703,14 @@ class dirstate:
             # In largefiles or readonly context
             mtime_boundary = None
 
-        if use_rust:
+        if self.use_rust_status(subrepos):
             try:
                 res = self._rust_status(
-                    match, listclean, listignored, listunknown
+                    match,
+                    listclean,
+                    listignored,
+                    listunknown,
+                    empty_dirs_keep_files,
                 )
                 return res + (mtime_boundary,)
             except rustmod.FallbackError:
@@ -1749,20 +1788,34 @@ class dirstate:
                         ladd(fn)
                     else:
                         madd(fn)
-                elif not t.mtime_likely_equal_to(timestamp.mtime_of(st)):
-                    # There might be a change in the future if for example the
-                    # internal clock is off, but this is a case where the issues
-                    # the user would face would be a lot worse and there is
-                    # nothing we can really do.
-                    ladd(fn)
-                elif listclean:
-                    cadd(fn)
+                else:
+                    reliable = None
+                    if mtime_boundary is not None:
+                        reliable = timestamp.reliable_mtime_of(
+                            st, mtime_boundary
+                        )
+                    elif t.mtime_likely_equal_to(timestamp.mtime_of(st)):
+                        # We can't compute the current fs time, so we're in
+                        # a readonly fs or a LFS context.
+                        cadd(fn)
+                        continue
+
+                    if reliable is None or not t.mtime_likely_equal_to(
+                        reliable
+                    ):
+                        # There might be a change in the future if for example
+                        # the internal clock is off, but this is a case where
+                        # the issues the user would face would be a lot worse
+                        # and there is nothing we can really do.
+                        ladd(fn)
+                    elif listclean:
+                        cadd(fn)
         status = scmutil.status(
             modified, added, removed, deleted, unknown, ignored, clean
         )
         return (lookup, status, mtime_boundary)
 
-    def matches(self, match):
+    def matches(self, match: MatcherT) -> Iterable[bytes]:
         """
         return files in the dirstate (in whatever state) filtered by match
         """
@@ -1795,7 +1848,9 @@ class dirstate:
                 files.append(self._map.docket.data_filename())
         return tuple(files)
 
-    def verify(self, m1, m2, p1, narrow_matcher=None):
+    def verify(
+        self, m1, m2, p1: bytes, narrow_matcher: Any | None = None
+    ) -> Iterator[bytes]:
         """
         check the dirstate contents against the parent manifest and yield errors
         """
