@@ -1,11 +1,11 @@
 # dispatch.py - command dispatching for mercurial
 #
-# Copyright 2005-2007 Matt Mackall <mpm@selenic.com>
+# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
 #
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-from __future__ import absolute_import, print_function
+from __future__ import annotations
 
 import errno
 import getopt
@@ -16,15 +16,14 @@ import re
 import signal
 import sys
 import traceback
-
+import typing
 
 from .i18n import _
-from .pycompat import getattr
 
 from hgdemandimport import tracing
 
 from . import (
-    cmdutil,
+    cmd_impls,
     color,
     commands,
     demandimport,
@@ -33,90 +32,48 @@ from . import (
     extensions,
     fancyopts,
     help,
-    hg,
     hook,
-    localrepo,
+    initialization,
+    main_script,
     profiling,
     pycompat,
-    rcutil,
     registrar,
-    requirements as requirementsmod,
+    repo as repo_utils,
     scmutil,
+    tables,
     ui as uimod,
     util,
-    vfs,
 )
-
+from .main_script import (
+    cmd_finder,
+    options as args_util,
+    paths as local_util,
+)
+from .repo import (
+    factory as repo_factory,
+)
 from .utils import (
     procutil,
     stringutil,
 )
 
 
-class request(object):
-    def __init__(
-        self,
-        args,
-        ui=None,
-        repo=None,
-        fin=None,
-        fout=None,
-        ferr=None,
-        fmsg=None,
-        prereposetups=None,
-    ):
-        self.args = args
-        self.ui = ui
-        self.repo = repo
-
-        # input/output/error streams
-        self.fin = fin
-        self.fout = fout
-        self.ferr = ferr
-        # separate stream for status/error messages
-        self.fmsg = fmsg
-
-        # remember options pre-parsed by _earlyparseopts()
-        self.earlyoptions = {}
-
-        # reposetups which run before extensions, useful for chg to pre-fill
-        # low-level repo state (for example, changelog) before extensions.
-        self.prereposetups = prereposetups or []
-
-        # store the parsed and canonical command
-        self.canonical_command = None
-
-    def _runexithandlers(self):
-        exc = None
-        handlers = self.ui._exithandlers
-        try:
-            while handlers:
-                func, args, kwargs = handlers.pop()
-                try:
-                    func(*args, **kwargs)
-                except:  # re-raises below
-                    if exc is None:
-                        exc = sys.exc_info()[1]
-                    self.ui.warnnoi18n(b'error in exit handlers:\n')
-                    self.ui.traceback(force=True)
-        finally:
-            if exc is not None:
-                raise exc
+initialization.init()
 
 
 def _flushstdio(ui, err):
     status = None
     # In all cases we try to flush stdio streams.
-    if util.safehasattr(ui, b'fout'):
+    if hasattr(ui, 'fout'):
         assert ui is not None  # help pytype
         assert ui.fout is not None  # help pytype
         try:
             ui.fout.flush()
-        except IOError as e:
+        except OSError as e:
             err = e
             status = -1
 
-    if util.safehasattr(ui, b'ferr'):
+    if hasattr(ui, 'ferr'):
         assert ui is not None  # help pytype
         assert ui.ferr is not None  # help pytype
         try:
@@ -127,7 +84,7 @@ def _flushstdio(ui, err):
             ui.ferr.flush()
         # There's not much we can do about an I/O error here. So (possibly)
         # change the status code and move on.
-        except IOError:
+        except OSError:
             status = -1
 
     return status
@@ -138,7 +95,7 @@ def run():
     try:
         initstdio()
         with tracing.log('parse args into request'):
-            req = request(pycompat.sysargv[1:])
+            req = main_script.request(pycompat.sysargv[1:])
 
         status = dispatch(req)
         _silencestdio()
@@ -149,96 +106,83 @@ def run():
     sys.exit(status & 255)
 
 
-if pycompat.ispy3:
+def initstdio():
+    # stdio streams on Python 3 are io.TextIOWrapper instances proxying another
+    # buffer. These streams will normalize \n to \r\n by default. Mercurial's
+    # preferred mechanism for writing output (ui.write()) uses io.BufferedWriter
+    # instances, which write to the underlying stdio file descriptor in binary
+    # mode. ui.write() uses \n for line endings and no line ending normalization
+    # is attempted through this interface. This "just works," even if the system
+    # preferred line ending is not \n.
+    #
+    # But some parts of Mercurial (e.g. hooks) can still send data to sys.stdout
+    # and sys.stderr. They will inherit the line ending normalization settings,
+    # potentially causing e.g. \r\n to be emitted. Since emitting \n should
+    # "just work," here we change the sys.* streams to disable line ending
+    # normalization, ensuring compatibility with our ui type.
 
-    def initstdio():
-        # stdio streams on Python 3 are io.TextIOWrapper instances proxying another
-        # buffer. These streams will normalize \n to \r\n by default. Mercurial's
-        # preferred mechanism for writing output (ui.write()) uses io.BufferedWriter
-        # instances, which write to the underlying stdio file descriptor in binary
-        # mode. ui.write() uses \n for line endings and no line ending normalization
-        # is attempted through this interface. This "just works," even if the system
-        # preferred line ending is not \n.
-        #
-        # But some parts of Mercurial (e.g. hooks) can still send data to sys.stdout
-        # and sys.stderr. They will inherit the line ending normalization settings,
-        # potentially causing e.g. \r\n to be emitted. Since emitting \n should
-        # "just work," here we change the sys.* streams to disable line ending
-        # normalization, ensuring compatibility with our ui type.
+    if sys.stdout is not None:
+        # write_through is new in Python 3.7.
+        kwargs = {
+            "newline": "\n",
+            "line_buffering": bool(sys.stdout.line_buffering),
+        }
+        if hasattr(sys.stdout, "write_through"):
+            # pytype: disable=attribute-error
+            kwargs["write_through"] = sys.stdout.write_through
+            # pytype: enable=attribute-error
+        sys.stdout = io.TextIOWrapper(
+            sys.stdout.buffer, sys.stdout.encoding, sys.stdout.errors, **kwargs
+        )
 
-        if sys.stdout is not None:
-            # write_through is new in Python 3.7.
-            kwargs = {
-                "newline": "\n",
-                "line_buffering": sys.stdout.line_buffering,
-            }
-            if util.safehasattr(sys.stdout, "write_through"):
-                kwargs["write_through"] = sys.stdout.write_through
-            sys.stdout = io.TextIOWrapper(
-                sys.stdout.buffer,
-                sys.stdout.encoding,
-                sys.stdout.errors,
-                **kwargs
-            )
+    if sys.stderr is not None:
+        kwargs = {
+            "newline": "\n",
+            "line_buffering": bool(sys.stderr.line_buffering),
+        }
+        if hasattr(sys.stderr, "write_through"):
+            # pytype: disable=attribute-error
+            kwargs["write_through"] = sys.stderr.write_through
+            # pytype: enable=attribute-error
+        sys.stderr = io.TextIOWrapper(
+            sys.stderr.buffer, sys.stderr.encoding, sys.stderr.errors, **kwargs
+        )
 
-        if sys.stderr is not None:
-            kwargs = {
-                "newline": "\n",
-                "line_buffering": sys.stderr.line_buffering,
-            }
-            if util.safehasattr(sys.stderr, "write_through"):
-                kwargs["write_through"] = sys.stderr.write_through
-            sys.stderr = io.TextIOWrapper(
-                sys.stderr.buffer,
-                sys.stderr.encoding,
-                sys.stderr.errors,
-                **kwargs
-            )
-
-        if sys.stdin is not None:
-            # No write_through on read-only stream.
-            sys.stdin = io.TextIOWrapper(
-                sys.stdin.buffer,
-                sys.stdin.encoding,
-                sys.stdin.errors,
-                # None is universal newlines mode.
-                newline=None,
-                line_buffering=sys.stdin.line_buffering,
-            )
-
-    def _silencestdio():
-        for fp in (sys.stdout, sys.stderr):
-            if fp is None:
-                continue
-            # Check if the file is okay
-            try:
-                fp.flush()
-                continue
-            except IOError:
-                pass
-            # Otherwise mark it as closed to silence "Exception ignored in"
-            # message emitted by the interpreter finalizer.
-            try:
-                fp.close()
-            except IOError:
-                pass
+    if sys.stdin is not None:
+        # No write_through on read-only stream.
+        sys.stdin = io.TextIOWrapper(
+            sys.stdin.buffer,
+            sys.stdin.encoding,
+            sys.stdin.errors,
+            # None is universal newlines mode.
+            newline=None,
+            line_buffering=bool(sys.stdin.line_buffering),
+        )
 
 
-else:
-
-    def initstdio():
-        for fp in (sys.stdin, sys.stdout, sys.stderr):
-            procutil.setbinary(fp)
-
-    def _silencestdio():
-        pass
+def _silencestdio():
+    for fp in (sys.stdout, sys.stderr):
+        if fp is None:
+            continue
+        # Check if the file is okay
+        try:
+            fp.flush()
+            continue
+        except OSError:
+            pass
+        # Otherwise mark it as closed to silence "Exception ignored in"
+        # message emitted by the interpreter finalizer.
+        try:
+            fp.close()
+        except OSError:
+            pass
 
 
 def _formatargs(args):
     return b' '.join(procutil.shellquote(a) for a in args)
 
 
-def dispatch(req):
+def dispatch(req: main_script.request) -> int:
     """run the command specified in req.args; returns an integer status code"""
     err = None
     try:
@@ -248,13 +192,17 @@ def dispatch(req):
         status = -1
 
     ret = _flushstdio(req.ui, err)
-    if ret:
+    if ret and not status:
         status = ret
     return status
 
 
-def _rundispatch(req):
-    with tracing.log('dispatch._rundispatch'):
+def _rundispatch(req: main_script.request) -> int:
+    # XXX figure out how to align old tracing and new tracing
+    with (
+        tracing.log('dispatch._rundispatch'),
+        util.rust_tracing_span("dispatch._rundispatch"),
+    ):
         if req.ferr:
             ferr = req.ferr
         elif req.ui:
@@ -265,9 +213,11 @@ def _rundispatch(req):
         try:
             if not req.ui:
                 req.ui = uimod.ui.load()
-            req.earlyoptions.update(_earlyparseopts(req.ui, req.args))
+            early_opts = args_util.early_parse_opts(req.ui, req.args)
+            req.earlyoptions.update(early_opts)
             if req.earlyoptions[b'traceback']:
                 req.ui.setconfig(b'ui', b'traceback', b'on', b'--traceback')
+                encoding.enable_rust_backtrace()
 
             # set ui streams from the request
             if req.fin:
@@ -282,7 +232,7 @@ def _rundispatch(req):
             ferr.write(inst.format())
             return -1
 
-        msg = _formatargs(req.args)
+        formattedargs = _formatargs(req.args)
         starttime = util.timer()
         ret = 1  # default of Python exit code on unhandled exception
         try:
@@ -303,9 +253,8 @@ def _rundispatch(req):
                 # maybe pager would quit without consuming all the output, and
                 # SIGPIPE was raised. we cannot print anything in this case.
                 pass
-            except IOError as inst:
-                if inst.errno != errno.EPIPE:
-                    raise
+            except BrokenPipeError:
+                pass
             ret = -1
         finally:
             duration = util.timer() - starttime
@@ -315,13 +264,13 @@ def _rundispatch(req):
                 req.ui.log(
                     b'uiblocked',
                     b'ui blocked ms\n',
-                    **pycompat.strkwargs(req.ui._blockedtimes)
+                    **pycompat.strkwargs(req.ui._blockedtimes),
                 )
             return_code = ret & 255
             req.ui.log(
                 b"commandfinish",
                 b"%s exited %d after %0.2f seconds\n",
-                msg,
+                formattedargs,
                 return_code,
                 duration,
                 return_code=return_code,
@@ -345,7 +294,7 @@ def _runcatch(req):
 
         ui = req.ui
         try:
-            for name in b'SIGBREAK', b'SIGHUP', b'SIGTERM':
+            for name in 'SIGBREAK', 'SIGHUP', 'SIGTERM':
                 num = getattr(signal, name, None)
                 if num:
                     signal.signal(num, catchterm)
@@ -356,10 +305,14 @@ def _runcatch(req):
             realcmd = None
             try:
                 cmdargs = fancyopts.fancyopts(
-                    req.args[:], commands.globalopts, {}
+                    req.args[:], cmd_impls.global_opts, {}
                 )
                 cmd = cmdargs[0]
-                aliases, entry = cmdutil.findcmd(cmd, commands.table, False)
+                aliases, entry = cmd_finder.find_cmd(
+                    cmd,
+                    tables.command_table,
+                    False,
+                )
                 realcmd = aliases[0]
             except (
                 error.UnknownCommand,
@@ -381,12 +334,18 @@ def _runcatch(req):
                 # shenanigans wherein a user does something like pass
                 # --debugger or --config=ui.debugger=1 as a repo
                 # name. This used to actually run the debugger.
+                nbargs = 4
+                hashiddenaccess = b'--hidden' in cmdargs
+                if hashiddenaccess:
+                    nbargs += 1
                 if (
-                    len(req.args) != 4
+                    len(req.args) != nbargs
                     or req.args[0] != b'-R'
                     or req.args[1].startswith(b'--')
                     or req.args[2] != b'serve'
                     or req.args[3] != b'--stdio'
+                    or hashiddenaccess
+                    and req.args[4] != b'--hidden'
                 ):
                     raise error.Abort(
                         _(b'potentially unsafe serve --stdio invocation: %s')
@@ -398,13 +357,24 @@ def _runcatch(req):
                 debugtrace = {b'pdb': pdb.set_trace}
                 debugmortem = {b'pdb': pdb.post_mortem}
 
-                # read --config before doing anything else
-                # (e.g. to change trust settings for reading .hg/hgrc)
-                cfgs = _parseconfig(req.ui, req.earlyoptions[b'config'])
+                # read --config-file and --config before doing anything else
+                # (e.g. to change trust settings for reading .hg/hgrc).
+
+                # cmdargs may not have been initialized here (in the case of an
+                # error), so use pycompat.sysargv instead.
+                file_cfgs = args_util.parse_config_files_opts(
+                    req.ui, pycompat.sysargv, req.earlyoptions[b'config_file']
+                )
+                cfgs = args_util.parse_config_opts(
+                    req.ui, req.earlyoptions[b'config']
+                )
 
                 if req.repo:
                     # copy configs that were passed on the cmdline (--config) to
                     # the repo ui
+                    for sec, name, val, source in file_cfgs:
+                        req.repo.ui.setconfig(sec, name, val, source=source)
+
                     for sec, name, val in cfgs:
                         req.repo.ui.setconfig(
                             sec, name, val, source=b'--config'
@@ -422,7 +392,7 @@ def _runcatch(req):
                     # debugging has been requested
                     with demandimport.deactivated():
                         try:
-                            debugmod = __import__(debugger)
+                            debugmod = __import__(pycompat.sysstr(debugger))
                         except ImportError:
                             pass  # Leave debugmod = pdb
 
@@ -495,9 +465,7 @@ def _callcatch(ui, func):
         try:
             # check if the command is in a disabled extension
             # (but don't check for extensions themselves)
-            formatted = help.formattedhelp(
-                ui, commands, inst.command, unknowncmd=True
-            )
+            formatted = help.formattedhelp(ui, inst.command, unknowncmd=True)
             ui.warn(nocmdmsg)
             ui.write(formatted)
         except (error.UnknownCommand, error.Abort):
@@ -511,7 +479,7 @@ def _callcatch(ui, func):
             if not suggested:
                 ui.warn(nocmdmsg)
                 ui.warn(_(b"(use 'hg help' for a list of commands)\n"))
-    except IOError:
+    except OSError:
         raise
     except KeyboardInterrupt:
         raise
@@ -528,7 +496,7 @@ def _callcatch(ui, func):
 def aliasargs(fn, givenargs):
     args = []
     # only care about alias 'args', ignore 'args' set by extensions.wrapfunction
-    if not util.safehasattr(fn, b'_origfunc'):
+    if not hasattr(fn, '_origfunc'):
         args = getattr(fn, 'args', args)
     if args:
         cmd = b' '.join(map(procutil.shellquote, args))
@@ -570,7 +538,7 @@ def aliasinterpolate(name, args, cmd):
     return r.sub(lambda x: replacemap[x.group()], cmd)
 
 
-class cmdalias(object):
+class cmdalias:
     def __init__(self, ui, name, definition, cmdtable, source):
         self.name = self.cmd = name
         self.cmdname = b''
@@ -584,8 +552,8 @@ class cmdalias(object):
         self.source = source
 
         try:
-            aliases, entry = cmdutil.findcmd(self.name, cmdtable)
-            for alias, e in pycompat.iteritems(cmdtable):
+            aliases, entry = cmd_finder.find_cmd(self.name, cmdtable)
+            for alias, e in cmdtable.items():
                 if e is entry:
                     self.cmd = alias
                     break
@@ -647,7 +615,7 @@ class cmdalias(object):
         self.givenargs = args
 
         try:
-            tableentry = cmdutil.findcmd(cmd, cmdtable, False)[1]
+            tableentry = cmd_finder.find_cmd(cmd, cmdtable, False)[1]
             if len(tableentry) > 2:
                 self.fn, self.opts, cmdhelp = tableentry
             else:
@@ -713,10 +681,11 @@ class cmdalias(object):
             'intents': set(),
             'optionalrepo': False,
             'inferrepo': False,
+            'need_dispatcher': False,
         }
         if name not in adefaults:
             raise AttributeError(name)
-        if self.badalias or util.safehasattr(self, b'shell'):
+        if self.badalias or hasattr(self, 'shell'):
             return adefaults[name]
         return getattr(self.fn, name)
 
@@ -742,7 +711,7 @@ class cmdalias(object):
             self.name,
             self.definition,
         )
-        if util.safehasattr(self, b'shell'):
+        if hasattr(self, 'shell'):
             return self.fn(ui, *args, **opts)
         else:
             try:
@@ -753,7 +722,7 @@ class cmdalias(object):
                 raise
 
 
-class lazyaliasentry(object):
+class lazyaliasentry:
     """like a typical command entry (func, opts, help), but is lazy"""
 
     def __init__(self, ui, name, definition, cmdtable, source):
@@ -811,14 +780,14 @@ def _parse(ui, args):
     cmdoptions = {}
 
     try:
-        args = fancyopts.fancyopts(args, commands.globalopts, options)
+        args = fancyopts.fancyopts(args, cmd_impls.global_opts, options)
     except getopt.GetoptError as inst:
         raise error.CommandError(None, stringutil.forcebytestr(inst))
 
     if args:
         cmd, args = args[0], args[1:]
-        aliases, entry = cmdutil.findcmd(
-            cmd, commands.table, ui.configbool(b"ui", b"strict")
+        aliases, entry = cmd_finder.find_cmd(
+            cmd, tables.command_table, ui.configbool(b"ui", b"strict")
         )
         cmd = aliases[0]
         args = aliasargs(entry[0], args)
@@ -833,9 +802,20 @@ def _parse(ui, args):
         cmd = None
         c = []
 
+    def global_opt_to_fancy_opt(opt_name: bytes) -> bytes:
+        # fancyopts() does this transform on `options`, but globalopts uses a
+        # '-', so that it is displayed in the help and accepted as input that
+        # way.
+        return opt_name.replace(b'-', b'_')
+
     # combine global options into local
-    for o in commands.globalopts:
-        c.append((o[0], o[1], options[o[1]], o[3]))
+    for o in cmd_impls.global_opts:
+        name = global_opt_to_fancy_opt(o[1])
+
+        # The fancyopts name is needed for `options`, but the original name
+        # needs to be used in the second element here, or the parsing for the
+        # command verb fails, saying the command has no such option.
+        c.append((o[0], o[1], options[name], o[3]))
 
     try:
         args = fancyopts.fancyopts(args, c, cmdoptions, gnu=True)
@@ -843,56 +823,25 @@ def _parse(ui, args):
         raise error.CommandError(cmd, stringutil.forcebytestr(inst))
 
     # separate global options back out
-    for o in commands.globalopts:
-        n = o[1]
+    for o in cmd_impls.global_opts:
+        n = global_opt_to_fancy_opt(o[1])
         options[n] = cmdoptions[n]
         del cmdoptions[n]
 
     return (cmd, cmd and entry[0] or None, args, options, cmdoptions)
 
 
-def _parseconfig(ui, config):
-    """parse the --config options from the command line"""
-    configs = []
-
-    for cfg in config:
-        try:
-            name, value = [cfgelem.strip() for cfgelem in cfg.split(b'=', 1)]
-            section, name = name.split(b'.', 1)
-            if not section or not name:
-                raise IndexError
-            ui.setconfig(section, name, value, b'--config')
-            configs.append((section, name, value))
-        except (IndexError, ValueError):
-            raise error.InputError(
-                _(
-                    b'malformed --config option: %r '
-                    b'(use --config section.name=value)'
-                )
-                % pycompat.bytestr(cfg)
-            )
-
-    return configs
-
-
-def _earlyparseopts(ui, args):
-    options = {}
-    fancyopts.fancyopts(
-        args,
-        commands.globalopts,
-        options,
-        gnu=not ui.plain(b'strictflags'),
-        early=True,
-        optaliases={b'repository': [b'repo']},
-    )
-    return options
-
-
 def _earlysplitopts(args):
     """Split args into a list of possible early options and remainder args"""
     shortoptions = b'R:'
     # TODO: perhaps 'debugger' should be included
-    longoptions = [b'cwd=', b'repository=', b'repo=', b'config=']
+    longoptions = [
+        b'cwd=',
+        b'repository=',
+        b'repo=',
+        b'config=',
+        b'config-file=',
+    ]
     return fancyopts.earlygetopt(
         args, shortoptions, longoptions, gnu=True, keepsep=True
     )
@@ -937,90 +886,31 @@ def runcommand(lui, repo, cmd, fullargs, ui, options, d, cmdpats, cmdoptions):
     return ret
 
 
-def _readsharedsourceconfig(ui, path):
-    """if the current repository is shared one, this tries to read
-    .hg/hgrc of shared source if we are in share-safe mode
-
-    Config read is loaded into the ui object passed
-
-    This should be called before reading .hg/hgrc or the main repo
-    as that overrides config set in shared source"""
-    try:
-        with open(os.path.join(path, b".hg", b"requires"), "rb") as fp:
-            requirements = set(fp.read().splitlines())
-            if not (
-                requirementsmod.SHARESAFE_REQUIREMENT in requirements
-                and requirementsmod.SHARED_REQUIREMENT in requirements
-            ):
-                return
-            hgvfs = vfs.vfs(os.path.join(path, b".hg"))
-            sharedvfs = localrepo._getsharedvfs(hgvfs, requirements)
-            root = sharedvfs.base
-            ui.readconfig(sharedvfs.join(b"hgrc"), root)
-    except IOError:
-        pass
-
-
-def _getlocal(ui, rpath, wd=None):
-    """Return (path, local ui object) for the given target path.
-
-    Takes paths in [cwd]/.hg/hgrc into account."
-    """
-    if wd is None:
-        try:
-            wd = encoding.getcwd()
-        except OSError as e:
-            raise error.Abort(
-                _(b"error getting current working directory: %s")
-                % encoding.strtolocal(e.strerror)
-            )
-
-    path = cmdutil.findrepo(wd) or b""
-    if not path:
-        lui = ui
-    else:
-        lui = ui.copy()
-        if rcutil.use_repo_hgrc():
-            _readsharedsourceconfig(lui, path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc"), path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc-not-shared"), path)
-
-    if rpath:
-        path = lui.expandpath(rpath)
-        lui = ui.copy()
-        if rcutil.use_repo_hgrc():
-            _readsharedsourceconfig(lui, path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc"), path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc-not-shared"), path)
-
-    return path, lui
-
-
 def _checkshellalias(lui, ui, args):
     """Return the function to run the shell alias, if it is required"""
     options = {}
 
     try:
-        args = fancyopts.fancyopts(args, commands.globalopts, options)
+        args = fancyopts.fancyopts(args, cmd_impls.global_opts, options)
     except getopt.GetoptError:
         return
 
     if not args:
         return
 
-    cmdtable = commands.table
+    cmdtable = tables.command_table
 
     cmd = args[0]
     try:
         strict = ui.configbool(b"ui", b"strict")
-        aliases, entry = cmdutil.findcmd(cmd, cmdtable, strict)
+        aliases, entry = cmd_finder.find_cmd(cmd, cmdtable, strict)
     except (error.AmbiguousCommand, error.UnknownCommand):
         return
 
     cmd = aliases[0]
     fn = entry[0]
 
-    if cmd and util.safehasattr(fn, b'shell'):
+    if cmd and hasattr(fn, 'shell'):
         # shell alias shouldn't receive early options which are consumed by hg
         _earlyopts, args = _earlysplitopts(args)
         d = lambda: fn(ui, *args[1:])
@@ -1030,16 +920,25 @@ def _checkshellalias(lui, ui, args):
 
 
 def _dispatch(req):
+    # check for cwd
+    old_cwd = None
+    cwd = req.earlyoptions[b'cwd']
+    try:
+        if cwd:
+            old_cwd = local_util.get_cwd()
+            os.chdir(cwd)
+        return _dispatch_post_cwd(req)
+    finally:
+        if old_cwd is not None:
+            os.chdir(old_cwd)
+
+
+def _dispatch_post_cwd(req):
     args = req.args
     ui = req.ui
 
-    # check for cwd
-    cwd = req.earlyoptions[b'cwd']
-    if cwd:
-        os.chdir(cwd)
-
     rpath = req.earlyoptions[b'repository']
-    path, lui = _getlocal(ui, rpath)
+    path, lui = local_util.get_local(ui, rpath)
 
     uis = {ui, lui}
 
@@ -1059,6 +958,16 @@ def _dispatch(req):
     if req.earlyoptions[b'profile']:
         for ui_ in uis:
             ui_.setconfig(b'profiling', b'enabled', b'true', b'--profile')
+    elif req.earlyoptions[b'profile'] is False:
+        # Check for it being set already, so that we don't pollute the config
+        # with this when using chg in the very common case that it's not
+        # enabled.
+        if lui.configbool(b'profiling', b'enabled'):
+            # Only do this on lui so that `chg foo` with a user config setting
+            # profiling.enabled=1 still shows profiling information (chg will
+            # specify `--no-profile` when `hg serve` is starting up, we don't
+            # want that to propagate to every later invocation).
+            lui.setconfig(b'profiling', b'enabled', b'false', b'--no-profile')
 
     profile = lui.configbool(b'profiling', b'enabled')
     with profiling.profile(lui, enabled=profile) as profiler:
@@ -1070,9 +979,9 @@ def _dispatch(req):
 
         # (uisetup and extsetup are handled in extensions.loadall)
 
-        # (reposetup is handled in hg.repository)
+        # (reposetup is handled in repo.factory.repository)
 
-        addaliases(lui, commands.table)
+        addaliases(lui, tables.command_table)
 
         # All aliases and commands are completely defined, now.
         # Check abbreviation/ambiguity of shell alias.
@@ -1089,13 +998,37 @@ def _dispatch(req):
             encoding.fallbackencoding = fallback
 
         fullargs = args
-        cmd, func, args, options, cmdoptions = _parse(lui, args)
+        try:
+            cmd, func, args, options, cmdoptions = _parse(lui, args)
+        except error.CommandError as e:
+            cause = e.__context__
+            if isinstance(cause, getopt.GetoptError):
+                if cause.opt and "config".startswith(cause.opt):
+                    # pycompat._getoptbwrapper() decodes bytes with latin-1
+                    opt = cause.opt.encode('latin-1')
+
+                    # pytype seems to merge all the types in each tuple index,
+                    # but o[1] is always bytes.
+                    all_long = typing.cast(
+                        set[bytes], {o[1] for o in cmd_impls.global_opts}
+                    )
+                    possible = [o for o in all_long if o.startswith(opt)]
+
+                    if len(possible) != 1:
+                        raise error.InputError(
+                            _(b"option --config may not be abbreviated")
+                        )
+            raise
 
         # store the canonical command name in request object for later access
         req.canonical_command = cmd
 
         if options[b"config"] != req.earlyoptions[b"config"]:
             raise error.InputError(_(b"option --config may not be abbreviated"))
+        if options[b"config_file"] != req.earlyoptions[b"config_file"]:
+            raise error.InputError(
+                _(b"option --config-file may not be abbreviated")
+            )
         if options[b"cwd"] != req.earlyoptions[b"cwd"]:
             raise error.InputError(_(b"option --cwd may not be abbreviated"))
         if options[b"repository"] != req.earlyoptions[b"repository"]:
@@ -1195,7 +1128,7 @@ def _dispatch(req):
         assert func is not None  # help out pytype
         if not func.norepo:
             # use the repo from the request only if we don't have -R
-            if not rpath and not cwd:
+            if not rpath and not req.earlyoptions[b'cwd']:
                 repo = req.repo
 
             if repo:
@@ -1206,7 +1139,7 @@ def _dispatch(req):
                 repo.ui.fmsg = ui.fmsg
             else:
                 try:
-                    repo = hg.repository(
+                    repo = repo_factory.repository(
                         ui,
                         path=path,
                         presetupfuncs=req.prereposetups,
@@ -1227,7 +1160,7 @@ def _dispatch(req):
                     if not func.optionalrepo:
                         if func.inferrepo and args and not path:
                             # try to infer -R from command args
-                            repos = pycompat.maplist(cmdutil.findrepo, args)
+                            repos = pycompat.maplist(repo_utils.find_repo, args)
                             guess = repos[0]
                             if guess and repos.count(guess) == len(repos):
                                 req.args = [b'--repository', guess] + fullargs
@@ -1253,6 +1186,8 @@ def _dispatch(req):
         msg = _formatargs(fullargs)
         ui.log(b"command", b'%s\n', msg)
         strcmdopt = pycompat.strkwargs(cmdoptions)
+        if func.need_dispatcher:
+            strcmdopt["__dispatch__"] = dispatch
         d = lambda: util.checksignature(func)(ui, *args, **strcmdopt)
         try:
             return runcommand(
@@ -1266,7 +1201,8 @@ def _dispatch(req):
 def _runcommand(ui, options, cmd, cmdfunc):
     """Run a command function, possibly with profiling enabled."""
     try:
-        with tracing.log("Running %s command" % cmd):
+        msg = "Running %s command" % cmd
+        with tracing.log(msg), util.rust_tracing_span(msg):
             return cmdfunc()
     except error.SignatureError:
         raise error.CommandError(cmd, _(b'invalid arguments'))
@@ -1327,7 +1263,9 @@ def _exceptionwarning(ui):
     else:
         bugtracker = ui.config(b'ui', b'supportcontact')
         if bugtracker is None:
-            bugtracker = _(b"https://mercurial-scm.org/wiki/BugTracker")
+            bugtracker = _(
+                b"https://foss.heptapod.net/mercurial/mercurial-devel/-/issues"
+            )
         warning = (
             _(
                 b"** unknown exception encountered, "
