@@ -1,21 +1,27 @@
 # mdiff.py - diff and patch routines for mercurial
 #
-# Copyright 2005, 2006 Matt Mackall <mpm@selenic.com>
+# Copyright 2005, 2006 Olivia Mackall <olivia@selenic.com>
 #
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-from __future__ import absolute_import
+from __future__ import annotations
 
 import re
 import struct
+import typing
 import zlib
 
-from .i18n import _
-from .pycompat import (
-    getattr,
-    setattr,
+from typing import (
+    Callable,
+    Iterable,
+    Iterator,
+    Optional,
+    Sequence,
+    cast,
 )
+
+from .i18n import _
 from . import (
     diffhelper,
     encoding,
@@ -24,21 +30,54 @@ from . import (
     pycompat,
     util,
 )
+from .interfaces import (
+    modules as intmod,
+)
+
 from .utils import dateutil
 
-bdiff = policy.importmod('bdiff')
-mpatch = policy.importmod('mpatch')
+bdiff: intmod.BDiff = policy.importmod('bdiff')
+rust_diff = policy.importrust('diff')
+mpatch: intmod.MPatch = policy.importmod('mpatch')
 
 blocks = bdiff.blocks
 fixws = bdiff.fixws
 patches = mpatch.patches
 patchedsize = mpatch.patchedsize
-textdiff = bdiff.bdiff
 splitnewlines = bdiff.splitnewlines
+# produce the historic generic diff
+textdiff = bdiff.bdiff
+text_diff = textdiff
+# produce diff using line as boundary
+# diff algorithm optimized for the manifest must be line aligned
+if rust_diff is not None:
+    line_diff = rust_diff.line_diff
+    manifest_diff = rust_diff.manifest_diff
+    storage_diff = rust_diff.line_diff
+else:
+    line_diff = bdiff.line_diff
+    manifest_diff = bdiff.manifest_diff
+    storage_diff = bdiff.storage_diff
+# diff algorithm optimized for storage purpose, might not be line aligned
+
+if typing.TYPE_CHECKING:
+    HunkLines = list[bytes]
+    """Lines of a hunk- a header, followed by line additions and deletions."""
+
+    HunkRange = tuple[int, int, int, int]
+    """HunkRange represents the range information of a hunk.
+
+    The tuple (s1, l1, s2, l2) forms the header '@@ -s1,l1 +s2,l2 @@'."""
+
+    Range = tuple[int, int]
+    """A (lowerbound, upperbound) range tuple."""
+
+    TypedBlock = tuple[intmod.BDiffBlock, bytes]
+    """A bdiff block with its type."""
 
 
 # TODO: this looks like it could be an attrs, which might help pytype
-class diffopts(object):
+class diffopts:
     """context is the number of context lines
     text treats all files as text
     showfunc enables diff -p output
@@ -79,31 +118,39 @@ class diffopts(object):
             v = opts.get(k)
             if v is None:
                 v = self.defaults[k]
-            setattr(self, k, v)
+            setattr(self, pycompat.sysstr(k), v)
 
         try:
             self.context = int(self.context)
         except ValueError:
-            raise error.Abort(
+            raise error.InputError(
                 _(b'diff context lines count must be an integer, not %r')
                 % pycompat.bytestr(self.context)
             )
 
     def copy(self, **kwargs):
-        opts = {k: getattr(self, k) for k in self.defaults}
+        opts = {k: getattr(self, pycompat.sysstr(k)) for k in self.defaults}
         opts = pycompat.strkwargs(opts)
         opts.update(kwargs)
         return diffopts(**opts)
 
+    def __bytes__(self):
+        return b", ".join(
+            b"%s: %r" % (k, getattr(self, pycompat.sysstr(k)))
+            for k in self.defaults
+        )
+
+    __str__ = encoding.strmethod(__bytes__)
+
 
 defaultopts = diffopts()
 
 
-def wsclean(opts, text, blank=True):
+def wsclean(opts: diffopts, text: bytes, blank: bool = True) -> bytes:
     if opts.ignorews:
-        text = bdiff.fixws(text, 1)
+        text = bdiff.fixws(text, True)
     elif opts.ignorewsamount:
-        text = bdiff.fixws(text, 0)
+        text = bdiff.fixws(text, False)
     if blank and opts.ignoreblanklines:
         text = re.sub(b'\n+', b'\n', text).strip(b'\n')
     if opts.ignorewseol:
@@ -111,7 +158,13 @@ def wsclean(opts, text, blank=True):
     return text
 
 
-def splitblock(base1, lines1, base2, lines2, opts):
+def splitblock(
+    base1: int,
+    lines1: Iterable[bytes],
+    base2: int,
+    lines2: Iterable[bytes],
+    opts: diffopts,
+) -> Iterable[TypedBlock]:
     # The input lines matches except for interwoven blank lines. We
     # transform it into a sequence of matching blocks and blank blocks.
     lines1 = [(wsclean(opts, l) and 1 or 0) for l in lines1]
@@ -132,12 +185,12 @@ def splitblock(base1, lines1, base2, lines2, opts):
             while i1 < e1 and lines1[i1] == 1 and lines2[i2] == 1:
                 i1 += 1
                 i2 += 1
-        yield [base1 + s1, base1 + i1, base2 + s2, base2 + i2], btype
+        yield (base1 + s1, base1 + i1, base2 + s2, base2 + i2), btype
         s1 = i1
         s2 = i2
 
 
-def hunkinrange(hunk, linerange):
+def hunkinrange(hunk: tuple[int, int], linerange: Range) -> bool:
     """Return True if `hunk` defined as (start, length) is in `linerange`
     defined as (lowerbound, upperbound).
 
@@ -163,7 +216,9 @@ def hunkinrange(hunk, linerange):
     return lowerbound < start + length and start < upperbound
 
 
-def blocksinrange(blocks, rangeb):
+def blocksinrange(
+    blocks: Iterable[TypedBlock], rangeb: Range
+) -> tuple[list[TypedBlock], Range]:
     """filter `blocks` like (a1, a2, b1, b2) from items outside line range
     `rangeb` from ``(b1, b2)`` point of view.
 
@@ -203,18 +258,24 @@ def blocksinrange(blocks, rangeb):
     return filteredblocks, (lba, uba)
 
 
-def chooseblocksfunc(opts=None):
+def chooseblocksfunc(opts: diffopts | None = None) -> intmod.BDiffBlocksFnc:
     if (
         opts is None
         or not opts.xdiff
-        or not util.safehasattr(bdiff, b'xdiffblocks')
+        or not getattr(bdiff, 'xdiffblocks', None)
     ):
         return bdiff.blocks
     else:
         return bdiff.xdiffblocks
 
 
-def allblocks(text1, text2, opts=None, lines1=None, lines2=None):
+def allblocks(
+    text1: bytes,
+    text2: bytes,
+    opts: diffopts | None = None,
+    lines1: Sequence[bytes] | None = None,
+    lines2: Sequence[bytes] | None = None,
+) -> Iterable[TypedBlock]:
     """Return (block, type) tuples, where block is an mdiff.blocks
     line entry. type is '=' for blocks matching exactly one another
     (bdiff blocks), '!' for non-matching blocks and '~' for blocks
@@ -236,8 +297,8 @@ def allblocks(text1, text2, opts=None, lines1=None, lines2=None):
         if i > 0:
             s = diff[i - 1]
         else:
-            s = [0, 0, 0, 0]
-        s = [s[1], s1[0], s[3], s1[2]]
+            s = (0, 0, 0, 0)
+        s = (s[1], s1[0], s[3], s1[2])
 
         # bdiff sometimes gives huge matches past eof, this check eats them,
         # and deals with the special first match case described above
@@ -256,7 +317,16 @@ def allblocks(text1, text2, opts=None, lines1=None, lines2=None):
         yield s1, b'='
 
 
-def unidiff(a, ad, b, bd, fn1, fn2, binary, opts=defaultopts):
+def unidiff(
+    a: bytes,
+    ad: bytes,
+    b: bytes,
+    bd: bytes,
+    fn1: bytes,
+    fn2: bytes,
+    binary: bool,
+    opts: diffopts = defaultopts,
+) -> tuple[list[bytes], Iterable[tuple[HunkRange | None, HunkLines]]]:
     """Return a unified diff as a (headers, hunks) tuple.
 
     If the diff is not null, `headers` is a list with unified diff header
@@ -267,7 +337,7 @@ def unidiff(a, ad, b, bd, fn1, fn2, binary, opts=defaultopts):
     Set binary=True if either a or b should be taken as a binary file.
     """
 
-    def datetag(date, fn=None):
+    def datetag(date: bytes, fn: bytes | None = None):
         if not opts.git and not opts.nodates:
             return b'\t%s' % date
         if fn and b' ' in fn:
@@ -336,10 +406,16 @@ def unidiff(a, ad, b, bd, fn1, fn2, binary, opts=defaultopts):
             b"+++ %s%s%s" % (bprefix, fn2, datetag(bd, fn2)),
         ]
 
-    return headerlines, hunks
+    # The possible bool is consumed from the iterator above in the `next()`
+    # call.
+    return headerlines, cast(
+        "Iterable[tuple[Optional[HunkRange], HunkLines]]", hunks
+    )
 
 
-def _unidiff(t1, t2, opts=defaultopts):
+def _unidiff(
+    t1: bytes, t2: bytes, opts: diffopts = defaultopts
+) -> Iterator[bool | tuple[HunkRange, HunkLines]]:
     """Yield hunks of a headerless unified diff from t1 and t2 texts.
 
     Each hunk consists of a (hunkrange, hunklines) tuple where `hunkrange` is a
@@ -367,7 +443,9 @@ def _unidiff(t1, t2, opts=defaultopts):
 
     lastfunc = [0, b'']
 
-    def yieldhunk(hunk):
+    def yieldhunk(
+        hunk: tuple[int, int, int, int, list[bytes]]
+    ) -> Iterable[tuple[HunkRange, HunkLines]]:
         (astart, a2, bstart, b2, delta) = hunk
         aend = contextend(a2, len(l1))
         alen = aend - astart
@@ -378,9 +456,10 @@ def _unidiff(t1, t2, opts=defaultopts):
             lastpos, func = lastfunc
             # walk backwards from the start of the context up to the start of
             # the previous hunk context until we find a line starting with an
-            # alphanumeric char.
-            for i in pycompat.xrange(astart - 1, lastpos - 1, -1):
-                if l1[i][0:1].isalnum():
+            # `identifier' char (alphabetic, _, or $, like GNU diffutils,
+            # libxdiff, and git).
+            for i in range(astart - 1, lastpos - 1, -1):
+                if l1[i][0:1].isalpha() or l1[i][0:1] in (b'_', b'$'):
                     func = b' ' + l1[i].rstrip()
                     # split long function name if ASCII. otherwise we have no
                     # idea where the multi-byte boundary is, so just leave it.
@@ -403,7 +482,7 @@ def _unidiff(t1, t2, opts=defaultopts):
         hunklines = (
             [b"@@ -%d,%d +%d,%d @@%s\n" % (hunkrange + (func,))]
             + delta
-            + [b' ' + l1[x] for x in pycompat.xrange(a2, aend)]
+            + [b' ' + l1[x] for x in range(a2, aend)]
         )
         # If either file ends without a newline and the last line of
         # that file is part of a hunk, a marker is printed. If the
@@ -412,7 +491,7 @@ def _unidiff(t1, t2, opts=defaultopts):
         # which the hunk can end in a shared line without a newline.
         skip = False
         if not t1.endswith(b'\n') and astart + alen == len(l1) + 1:
-            for i in pycompat.xrange(len(hunklines) - 1, -1, -1):
+            for i in range(len(hunklines) - 1, -1, -1):
                 if hunklines[i].startswith((b'-', b' ')):
                     if hunklines[i].startswith(b' '):
                         skip = True
@@ -420,7 +499,7 @@ def _unidiff(t1, t2, opts=defaultopts):
                     hunklines.insert(i + 1, diffhelper.MISSING_NEWLINE_MARKER)
                     break
         if not skip and not t2.endswith(b'\n') and bstart + blen == len(l2) + 1:
-            for i in pycompat.xrange(len(hunklines) - 1, -1, -1):
+            for i in range(len(hunklines) - 1, -1, -1):
                 if hunklines[i].startswith(b'+'):
                     hunklines[i] += b'\n'
                     hunklines.insert(i + 1, diffhelper.MISSING_NEWLINE_MARKER)
@@ -463,16 +542,14 @@ def _unidiff(t1, t2, opts=defaultopts):
                 if not has_hunks:
                     has_hunks = True
                     yield True
-                for x in yieldhunk(hunk):
-                    yield x
+                yield from yieldhunk(hunk)
         if prev:
             # we've joined the previous hunk, record the new ending points.
-            hunk[1] = a2
-            hunk[3] = b2
+            hunk = (hunk[0], a2, hunk[2], b2, hunk[4])
             delta = hunk[4]
         else:
             # create a new hunk
-            hunk = [astart, a2, bstart, b2, delta]
+            hunk = (astart, a2, bstart, b2, delta)
 
         delta[len(delta) :] = [b' ' + x for x in l1[astart:a1]]
         delta[len(delta) :] = [b'-' + x for x in old]
@@ -482,13 +559,12 @@ def _unidiff(t1, t2, opts=defaultopts):
         if not has_hunks:
             has_hunks = True
             yield True
-        for x in yieldhunk(hunk):
-            yield x
+        yield from yieldhunk(hunk)
     elif not has_hunks:
         yield False
 
 
-def b85diff(to, tn):
+def b85diff(to: bytes | None, tn: bytes | None) -> bytes:
     '''print base85-encoded binary diff'''
 
     def fmtline(line):
@@ -525,11 +601,14 @@ def b85diff(to, tn):
     return b''.join(ret)
 
 
-def patchtext(bin):
+DIFF_HEADER = struct.Struct(b">lll")
+
+
+def patchtext(bin: bytes) -> bytes:
     pos = 0
     t = []
     while pos < len(bin):
-        p1, p2, l = struct.unpack(b">lll", bin[pos : pos + 12])
+        p1, p2, l = DIFF_HEADER.unpack(bin[pos : pos + 12])
         pos += 12
         t.append(bin[pos : pos + l])
         pos += l
@@ -544,13 +623,59 @@ def patch(a, bin):
 
 
 # similar to difflib.SequenceMatcher.get_matching_blocks
-def get_matching_blocks(a, b):
+def get_matching_blocks(a: bytes, b: bytes) -> list[tuple[int, int, int]]:
     return [(d[0], d[2], d[1] - d[0]) for d in bdiff.blocks(a, b)]
 
 
-def trivialdiffheader(length):
-    return struct.pack(b">lll", 0, 0, length) if length else b''
+def trivialdiffheader(length: int) -> bytes:
+    return DIFF_HEADER.pack(0, 0, length) if length else b''
 
 
-def replacediffheader(oldlen, newlen):
-    return struct.pack(b">lll", 0, oldlen, newlen)
+def replacediffheader(oldlen: int, newlen: int) -> bytes:
+    return DIFF_HEADER.pack(0, oldlen, newlen)
+
+
+def full_text_from_delta(
+    delta: bytes,
+    base_size: int,
+    base_func: Callable[[], bytes],
+) -> bytes:
+    """build a full text from a binary delta and its full text base
+
+    The base is provided as a callback as in some case we won't need it at all.
+    """
+    # special case deltas which replace entire base; no need to decode
+    # base revision. this neatly avoids censored bases, which throw when
+    # they're decoded.
+    if delta[: DIFF_HEADER.size] == replacediffheader(
+        base_size, len(delta) - DIFF_HEADER.size
+    ):
+        fulltext = delta[DIFF_HEADER.size :]
+    else:
+        basetext = base_func()
+        fulltext = patch(basetext, delta)
+    return fulltext
+
+
+def first_bytes(
+    delta: bytes,
+    size: int,
+    has_base: bool = True,
+) -> tuple[int, bytes, int]:
+    """
+    inherited-bytes, new-bytes, inherited-bytes
+    """
+    if len(delta) == 0:
+        if has_base:
+            return (size, b'', 0)
+        else:
+            return (0, b'', 0)
+    elif len(delta) < DIFF_HEADER.size:
+        return (0, delta[:size], 0)
+    o = DIFF_HEADER.size
+    start, old_size, new_size = DIFF_HEADER.unpack(delta[:o])
+    if start >= size:
+        return (size, b'', 0)
+    else:
+        cap_size = min(size - start, new_size - start)
+        return (start, delta[o : o + cap_size], size - start - cap_size)
