@@ -7,25 +7,41 @@
 # GNU General Public License version 2 or any later version.
 """Helper class to compute deltas stored inside revlogs"""
 
+from __future__ import annotations
 
+import enum
 import struct
+import typing
 
-from ..interfaces import repository
+from typing import (
+    Callable,
+    TypedDict,
+)
+
+from ..interfaces import (
+    repository,
+    revlog as revlog_t,
+)
 from .. import revlogutils
 
 ### Internal utily constants
 
-KIND_CHANGELOG = 1001  # over 256 to not be comparable with a bytes
-KIND_MANIFESTLOG = 1002
-KIND_FILELOG = 1003
-KIND_OTHER = 1004
 
-ALL_KINDS = {
-    KIND_CHANGELOG,
-    KIND_MANIFESTLOG,
-    KIND_FILELOG,
-    KIND_OTHER,
-}
+class Kind(enum.IntEnum):
+    # we picked number over 256 to not be comparable with a bytes
+    CHANGELOG = 1001
+    MANIFESTLOG = 1002
+    FILELOG = 1003
+    OTHER = 1004
+
+
+# Keep previous name for compatibility
+KIND_CHANGELOG: Kind = Kind.CHANGELOG
+KIND_MANIFESTLOG: Kind = Kind.MANIFESTLOG
+KIND_FILELOG: Kind = Kind.FILELOG
+KIND_OTHER: Kind = Kind.OTHER
+ALL_KINDS: set[Kind] = set(Kind)
+RevlogKindT = Kind
 
 ### Index entry key
 #
@@ -132,11 +148,20 @@ CHANGELOGV2 = 0xD34D
 FLAG_INLINE_DATA = 1 << 16
 # Only used by v1, implied by v2.
 FLAG_GENERALDELTA = 1 << 17
+
+# This filelog use a flag to signal metadata presence in a file revision
+FLAG_FILELOG_META = 1 << 18
+
+# snapshot delta should be explicitly marked as such
+FLAG_DELTA_INFO = 1 << 19
+
 REVLOG_DEFAULT_FLAGS = FLAG_INLINE_DATA
 REVLOG_DEFAULT_FORMAT = REVLOGV1
 REVLOG_DEFAULT_VERSION = REVLOG_DEFAULT_FORMAT | REVLOG_DEFAULT_FLAGS
 REVLOGV0_FLAGS = 0
-REVLOGV1_FLAGS = FLAG_INLINE_DATA | FLAG_GENERALDELTA
+REVLOGV1_FLAGS = (
+    FLAG_INLINE_DATA | FLAG_GENERALDELTA | FLAG_FILELOG_META | FLAG_DELTA_INFO
+)
 REVLOGV2_FLAGS = FLAG_INLINE_DATA
 CHANGELOGV2_FLAGS = 0
 
@@ -217,8 +242,20 @@ REVIDX_ISCENSORED = repository.REVISION_FLAG_CENSORED
 REVIDX_ELLIPSIS = repository.REVISION_FLAG_ELLIPSIS
 # revision data is stored externally
 REVIDX_EXTSTORED = repository.REVISION_FLAG_EXTSTORED
-# revision changes files in a way that could affect copy tracing.
+# changelog revision changes files in a way that could affect copy tracing.
 REVIDX_HASCOPIESINFO = repository.REVISION_FLAG_HASCOPIESINFO
+
+# filelog revision has metadata
+REVIDX_HASMETA = repository.FILEREVISION_FLAG_HASMETA
+
+# revision is stored as a snapshot-delta
+REVIDX_DELTA_IS_SNAPSHOT = repository.REVISION_FLAG_DELTA_IS_SNAPSHOT
+
+REVIDX_DELTA_QUALITY = repository.REVISION_FLAG_DELTA_HAS_QUALITY
+REVIDX_DELTA_GOOD = repository.REVISION_FLAG_DELTA_IS_GOOD
+REVIDX_DELTA_P1_SMALL = repository.REVISION_FLAG_DELTA_P1_IS_SMALL
+REVIDX_DELTA_P2_SMALL = repository.REVISION_FLAG_DELTA_P2_IS_SMALL
+
 REVIDX_DEFAULT_FLAGS = 0
 # stable order in which flags need to be processed and their processors applied
 REVIDX_FLAGS_ORDER = [
@@ -226,8 +263,27 @@ REVIDX_FLAGS_ORDER = [
     REVIDX_ELLIPSIS,
     REVIDX_EXTSTORED,
     REVIDX_HASCOPIESINFO,
+    REVIDX_HASMETA,
+    REVIDX_DELTA_IS_SNAPSHOT,
+    REVIDX_DELTA_QUALITY,
+    REVIDX_DELTA_GOOD,
+    REVIDX_DELTA_P1_SMALL,
+    REVIDX_DELTA_P2_SMALL,
 ]
 
+REVIDX_DELTA_INFO_FLAGS = (
+    REVIDX_DELTA_IS_SNAPSHOT
+    | REVIDX_DELTA_QUALITY
+    | REVIDX_DELTA_GOOD
+    | REVIDX_DELTA_P1_SMALL
+    | REVIDX_DELTA_P2_SMALL
+)
+
+# the flags that doesn't alter the revision content
+REVIDX_NEUTRAL_FLAGS = (
+    REVIDX_HASCOPIESINFO | REVIDX_HASMETA | REVIDX_DELTA_INFO_FLAGS
+)
+
 # bitmark for flags that could cause rawdata content change
 REVIDX_RAWTEXT_CHANGING_FLAGS = REVIDX_ISCENSORED | REVIDX_EXTSTORED
 
@@ -268,34 +324,54 @@ def _from_flag(flag):
     return lambda flags: bool(flags & flag)
 
 
-FEATURES_BY_VERSION = {
+if typing.TYPE_CHECKING:
+    _FromFlagsFnc = Callable[[int], bool]
+
+    class RevlogFeatures(TypedDict):
+        inline: _FromFlagsFnc
+        generaldelta: _FromFlagsFnc
+        hasmeta_flag: _FromFlagsFnc
+        sidedata: bool
+        docket: bool
+        delta_info: _FromFlagsFnc
+
+
+FEATURES_BY_VERSION: dict[int, RevlogFeatures] = {
     REVLOGV0: {
-        b'inline': _no,
-        b'generaldelta': _no,
-        b'sidedata': False,
-        b'docket': False,
+        'inline': _no,
+        'generaldelta': _no,
+        'hasmeta_flag': _no,
+        'delta_info': _no,
+        'sidedata': False,
+        'docket': False,
     },
     REVLOGV1: {
-        b'inline': _from_flag(FLAG_INLINE_DATA),
-        b'generaldelta': _from_flag(FLAG_GENERALDELTA),
-        b'sidedata': False,
-        b'docket': False,
+        'inline': _from_flag(FLAG_INLINE_DATA),
+        'generaldelta': _from_flag(FLAG_GENERALDELTA),
+        'hasmeta_flag': _from_flag(FLAG_FILELOG_META),
+        'delta_info': _from_flag(FLAG_DELTA_INFO),
+        'sidedata': False,
+        'docket': False,
     },
     REVLOGV2: {
         # The point of inline-revlog is to reduce the number of files used in
         # the store. Using a docket defeat this purpose. So we needs other
         # means to reduce the number of files for revlogv2.
-        b'inline': _no,
-        b'generaldelta': _yes,
-        b'sidedata': True,
-        b'docket': True,
+        'inline': _no,
+        'generaldelta': _yes,
+        'hasmeta_flag': _no,  # Should become yes at some point
+        'delta_info': _no,  # XXX we should make that True at some point
+        'sidedata': True,
+        'docket': True,
     },
     CHANGELOGV2: {
-        b'inline': _no,
+        'inline': _no,
         # General delta is useless for changelog since we don't do any delta
-        b'generaldelta': _no,
-        b'sidedata': True,
-        b'docket': True,
+        'generaldelta': _no,
+        'hasmeta_flag': _no,  # Should become yes at some point
+        'delta_info': _no,
+        'sidedata': True,
+        'docket': True,
     },
 }
 
@@ -304,15 +380,25 @@ SPARSE_REVLOG_MAX_CHAIN_LENGTH = 1000
 
 ### What should be done with a cached delta and its base ?
 
+DELTA_REUSE_NO = revlog_t.DeltaBaseReusePolicy.NO_DELTA
+
 # Ignore the cache when considering candidates.
 #
 # The cached delta might be used, but the delta base will not be scheduled for
 # usage earlier than in "normal" order.
-DELTA_BASE_REUSE_NO = 0
+DELTA_BASE_REUSE_NO = revlog_t.DeltaBaseReusePolicy.NO
 
 # Prioritize trying the cached delta base
 #
 # The delta base will be tested for validy first. So that the cached deltas get
 # used when possible.
-DELTA_BASE_REUSE_TRY = 1
-DELTA_BASE_REUSE_FORCE = 2
+DELTA_BASE_REUSE_TRY = revlog_t.DeltaBaseReusePolicy.TRY
+DELTA_BASE_REUSE_FORCE = revlog_t.DeltaBaseReusePolicy.FORCE
+
+FILELOG_HASMETA_UPGRADE = 1
+FILELOG_HASMETA_DOWNGRADE = 2
+
+# metadata marker in stored content
+# (not strickly revlog related, but used most for revlog)
+META_MARKER = b'\x01\n'
+META_MARKER_SIZE = len(META_MARKER)
