@@ -50,11 +50,9 @@ from . import (
     discovery,
     error,
     exchange,
-    hg,
     lock as lockmod,
     mdiff,
     merge,
-    mergestate as mergestatemod,
     patch,
     phases,
     pycompat,
@@ -64,6 +62,9 @@ from . import (
     util,
     vfs as vfsmod,
 )
+from .cmd_impls import (
+    update as up_impl,
+)
 from .utils import (
     dateutil,
     stringutil,
@@ -686,8 +687,8 @@ def _docreatecmd(ui, repo, pats, opts):
             with repo.dirstate.changing_parents(repo):
                 scmutil.movedirstate(repo, parent, match)
         else:
-            hg.update(repo, parent.node())
-            ms = mergestatemod.mergestate.read(repo)
+            up_impl.update(repo, parent.node())
+            ms = repo.mergestate()
             if not ms.unresolvedcount():
                 ms.reset()
 
@@ -874,7 +875,7 @@ def mergefiles(ui, repo, wctx, shelvectx) -> None:
     """updates to wctx and merges the changes from shelvectx into the
     dirstate."""
     with ui.configoverride({(b'ui', b'quiet'): True}):
-        hg.update(repo, wctx.node())
+        up_impl.update(repo, wctx.node())
         cmdutil.revert(ui, repo, shelvectx)
 
 
@@ -902,7 +903,7 @@ def unshelvecontinue(ui, repo, state: shelvedstate, opts) -> None:
     basename = state.name
     with repo.lock():
         checkparents(repo, state)
-        ms = mergestatemod.mergestate.read(repo)
+        ms = repo.mergestate()
         if ms.unresolvedcount():
             raise error.Abort(
                 _(b"unresolved conflicts, can't continue"),
