@@ -8,21 +8,39 @@
 from __future__ import annotations
 
 import random
+import typing
 
 from ..i18n import _
 from .. import (
     error,
-    localrepo,
     requirements,
     revlog,
     util,
 )
-
+from ..repo import (
+    creation,
+)
 from ..utils import compression
 
+if typing.TYPE_CHECKING:
+    from typing import (
+        ClassVar,
+        Final,
+        TypeVar,
+    )
+    from ..interfaces.types import (
+        RepoT,
+        RequirementSetT,
+        RequirementT,
+        UiT,
+    )
+
+    _FormatVariantTypeT = TypeVar(
+        '_FormatVariantTypeT', bound=type["formatvariant"]
+    )
 
 # list of requirements that request a clone of all revlog if added/removed
-RECLONES_REQUIREMENTS = {
+RECLONES_REQUIREMENTS: Final[RequirementSetT] = {
     requirements.CHANGELOGV2_REQUIREMENT,
     requirements.DELTA_INFO_REQUIREMENT,
     requirements.GENERALDELTA_REQUIREMENT,
@@ -31,7 +49,7 @@ RECLONES_REQUIREMENTS = {
 }
 
 
-def preservedrequirements(repo):
+def preservedrequirements(repo: RepoT) -> RequirementSetT:
     preserved = {
         requirements.NARROW_REQUIREMENT,
         requirements.SHARED_REQUIREMENT,
@@ -39,8 +57,8 @@ def preservedrequirements(repo):
     return preserved & repo.requirements
 
 
-FORMAT_VARIANT = b'deficiency'
-OPTIMISATION = b'optimization'
+FORMAT_VARIANT: Final[bytes] = b'deficiency'
+OPTIMISATION: Final[bytes] = b'optimization'
 
 
 class improvement:
@@ -104,7 +122,7 @@ class improvement:
 allformatvariant: list[type[formatvariant]] = []
 
 
-def registerformatvariant(cls):
+def registerformatvariant(cls: _FormatVariantTypeT) -> _FormatVariantTypeT:
     allformatvariant.append(cls)
     return cls
 
@@ -112,15 +130,15 @@ def registerformatvariant(cls):
 class formatvariant(improvement):
     """an improvement subclass dedicated to repository format"""
 
-    type = FORMAT_VARIANT
+    type: ClassVar[bytes] = FORMAT_VARIANT
 
     @staticmethod
-    def fromrepo(repo):
+    def fromrepo(repo: RepoT):
         """current value of the variant in the repository"""
         raise NotImplementedError()
 
     @staticmethod
-    def fromconfig(repo):
+    def fromconfig(repo: RepoT):
         """current value of the variant in the configuration"""
         raise NotImplementedError()
 
@@ -133,146 +151,176 @@ class requirementformatvariant(formatvariant):
     """
 
     # the requirement that control this format variant
-    _requirement = None
+    _requirement: ClassVar[RequirementT | None] = None
 
     @staticmethod
-    def _newreporequirements(ui):
-        return localrepo.newreporequirements(
-            ui, localrepo.defaultcreateopts(ui)
+    def _newreporequirements(ui: UiT) -> RequirementSetT:
+        return creation.new_repo_requirements(
+            ui,
+            creation.default_create_opts(ui),
         )
 
     @classmethod
-    def fromrepo(cls, repo):
+    def fromrepo(cls, repo: RepoT) -> bool:
         assert cls._requirement is not None
         return cls._requirement in repo.requirements
 
     @classmethod
-    def fromconfig(cls, repo):
+    def fromconfig(cls, repo: RepoT) -> bool:
         assert cls._requirement is not None
         return cls._requirement in cls._newreporequirements(repo.ui)
 
 
 @registerformatvariant
 class fncache(requirementformatvariant):
-    name = b'fncache'
+    name: ClassVar[bytes] = b'fncache'
 
-    _requirement = requirements.FNCACHE_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.FNCACHE_REQUIREMENT
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'long and reserved filenames may not work correctly; '
         b'repository performance is sub-optimal'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'repository will be more resilient to storing '
         b'certain paths and performance of certain '
         b'operations should be improved'
     )
 
 
+@registerformatvariant
+class file_index_v1(requirementformatvariant):
+    name: ClassVar[bytes] = b'fileindex-v1'
+
+    _requirement: ClassVar[RequirementT] = requirements.FILEINDEXV1_REQUIREMENT
+
+    default: ClassVar[bool] = False
+
+    description: ClassVar[bytes] = _(
+        b'replaces the fncache with a more efficient and powerful storage '
+        b'format'
+    )
+
+    upgrademessage: ClassVar[bytes] = _(
+        b'transactions that add files will be faster in large repos'
+    )
+
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
+    touches_dirstate: ClassVar[bool] = False
+    compatible_with_share: ClassVar[bool] = False
+
+
 @registerformatvariant
 class dirstatev2(requirementformatvariant):
-    name = b'dirstate-v2'
-    _requirement = requirements.DIRSTATE_V2_REQUIREMENT
+    name: ClassVar[bytes] = b'dirstate-v2'
+    _requirement: ClassVar[RequirementT] = requirements.DIRSTATE_V2_REQUIREMENT
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'version 1 of the dirstate file format requires '
         b'reading and parsing it all at once.\n'
         b'Version 2 has a better structure,'
         b'better information and lighter update mechanism'
     )
 
-    upgrademessage = _(b'"hg status" will be faster')
+    upgrademessage: ClassVar[bytes] = _(b'"hg status" will be faster')
 
-    touches_filelogs = False
-    touches_manifests = False
-    touches_changelog = False
-    touches_requirements = True
-    touches_dirstate = True
-    compatible_with_share = True
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
+    touches_dirstate: ClassVar[bool] = True
+    compatible_with_share: ClassVar[bool] = True
 
 
 @registerformatvariant
 class dirstatetrackedkey(requirementformatvariant):
-    name = b'tracked-hint'
-    _requirement = requirements.DIRSTATE_TRACKED_HINT_V1
+    name: ClassVar[bytes] = b'tracked-hint'
+    _requirement: ClassVar[RequirementT] = requirements.DIRSTATE_TRACKED_HINT_V1
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'Add a small file to help external tooling that watch the tracked set'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'external tools will be informated of potential change in the tracked set'
     )
 
-    touches_filelogs = False
-    touches_manifests = False
-    touches_changelog = False
-    touches_requirements = True
-    touches_dirstate = True
-    compatible_with_share = True
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
+    touches_dirstate: ClassVar[bool] = True
+    compatible_with_share: ClassVar[bool] = True
 
 
 @registerformatvariant
 class dotencode(requirementformatvariant):
-    name = b'dotencode'
+    name: ClassVar[bytes] = b'dotencode'
 
-    _requirement = requirements.DOTENCODE_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.DOTENCODE_REQUIREMENT
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'storage of filenames beginning with a period or '
         b'space may not work correctly'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'repository will be better able to store files '
         b'beginning with a space or period'
     )
 
-    touches_filelogs = True
-    touches_manifests = False
-    touches_changelog = False
-    touches_requirements = True
-    touches_dirstate = False
-    compatible_with_share = True
+    touches_filelogs: ClassVar[bool] = True
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
+    touches_dirstate: ClassVar[bool] = False
+    compatible_with_share: ClassVar[bool] = True
 
 
 @registerformatvariant
 class plain_encode(requirementformatvariant):
-    name = b'fragile-plain-encode'
+    name: ClassVar[bytes] = b'fragile-plain-encode'
 
-    _requirement = requirements.PLAIN_ENCODE_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.PLAIN_ENCODE_REQUIREMENT
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(b'using this option is dangerous and not recommended')
+    description: ClassVar[bytes] = _(
+        b'using this option is dangerous and not recommended'
+    )
 
-    upgrademessage = _(b'you will move to this dangerous format, are you sure?')
-    touches_filelogs = True
-    touches_manifests = False
-    touches_changelog = False
-    touches_requirements = True
-    touches_dirstate = False
-    compatible_with_share = True
+    upgrademessage: ClassVar[bytes] = _(
+        b'you will move to this dangerous format, are you sure?'
+    )
+    touches_filelogs: ClassVar[bool] = True
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
+    touches_dirstate: ClassVar[bool] = False
+    compatible_with_share: ClassVar[bool] = True
 
 
 @registerformatvariant
 class generaldelta(requirementformatvariant):
-    name = b'generaldelta'
+    name: ClassVar[bytes] = b'generaldelta'
 
-    _requirement = requirements.GENERALDELTA_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.GENERALDELTA_REQUIREMENT
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'deltas within internal storage are unable to '
         b'choose optimal revisions; repository is larger and '
         b'slower than it could be; interaction with other '
@@ -280,7 +328,7 @@ class generaldelta(requirementformatvariant):
         b'resources, making "hg push" and "hg pull" slower'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'repository storage will be able to create '
         b'optimal deltas; new repository data will be '
         b'smaller and read times should decrease; '
@@ -293,29 +341,29 @@ class generaldelta(requirementformatvariant):
 
 @registerformatvariant
 class sharesafe(requirementformatvariant):
-    name = b'share-safe'
-    _requirement = requirements.SHARESAFE_REQUIREMENT
+    name: ClassVar[bytes] = b'share-safe'
+    _requirement: ClassVar[RequirementT] = requirements.SHARESAFE_REQUIREMENT
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'old shared repositories do not share source repository '
         b'requirements and config. This leads to various problems '
         b'when the source repository format is upgraded or some new '
         b'extensions are enabled.'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'Upgrades a repository to share-safe format so that future '
         b'shares of this repository share its requirements and configs.'
     )
 
-    postdowngrademessage = _(
+    postdowngrademessage: ClassVar[bytes] = _(
         b'repository downgraded to not use share safe mode, '
         b'existing shares will not work and need to be reshared.'
     )
 
-    postupgrademessage = _(
+    postupgrademessage: ClassVar[bytes] = _(
         b'repository upgraded to share safe mode, existing'
         b' shares will still work in old non-safe mode. '
         b'Re-share existing shares to use them in safe mode'
@@ -323,161 +371,144 @@ class sharesafe(requirementformatvariant):
     )
 
     # upgrade only needs to change the requirements
-    touches_filelogs = False
-    touches_manifests = False
-    touches_changelog = False
-    touches_requirements = True
-
-
-@registerformatvariant
-class hasmeta_filelog(requirementformatvariant):
-    name = b'hasmeta_flag'
-
-    _requirement = requirements.FILELOG_METAFLAG_REQUIREMENT
-
-    default = False
-
-    description = _(
-        b'copy metadata in filelog uses a fragile marking and some cases '
-        b'require unpacking file content to determine if copy information '
-        b'exists.'
-    )
-
-    upgrademessage = _(
-        b'A dedicated index flag marks file revisions with copy information.'
-    )
-
-    # upgrade only needs to process the filelogs
-    touches_filelogs = True
-    touches_manifests = False
-    touches_changelog = False
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
+    touches_changelog: ClassVar[bool] = False
+    touches_requirements: ClassVar[bool] = True
 
 
 @registerformatvariant
 class sparserevlog(requirementformatvariant):
-    name = b'sparserevlog'
+    name: ClassVar[bytes] = b'sparserevlog'
 
-    _requirement = requirements.SPARSEREVLOG_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.SPARSEREVLOG_REQUIREMENT
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'in order to limit disk reading and memory usage on older '
-        b'version, the span of a delta chain from its root to its '
+        b'versions, the span of a delta chain from its root to its '
         b'end is limited, whatever the relevant data in this span. '
-        b'This can severly limit Mercurial ability to build good '
-        b'chain of delta resulting is much more storage space being '
-        b'taken and limit reusability of on disk delta during '
+        b"This can severely limit Mercurial's ability to build a good "
+        b'chain of delta resulting in much more storage space being '
+        b'taken, and limit reusability of on disk delta during '
         b'exchange.'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'Revlog supports delta chain with more unused data '
-        b'between payload. These gaps will be skipped at read '
+        b'between payloads. These gaps will be skipped at read '
         b'time. This allows for better delta chains, making a '
-        b'better compression and faster exchange with server.'
+        b'better compression and faster exchange with the server.'
     )
 
 
 @registerformatvariant
 class delta_info_flags(requirementformatvariant):
-    name = b'delta-info-flags'
+    name: ClassVar[bytes] = b'delta-info-flags'
 
-    _requirement = requirements.DELTA_INFO_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.DELTA_INFO_REQUIREMENT
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'Store information about stored delta in the index to help optimize '
-        b'local and remote delta chain.'
+        b'local and remote delta chain and stop using a fragile marking to '
+        b'signal copy metadata presence that requires copy metadata in '
+        b'filelog uses a fragile marking and some cases require unpacking file'
+        b' content to determine if copy information exists in some case.'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'Using this format will improve local storage, speedup '
         b'local operation and help to preserve storage quality and '
-        b'performace during exchanges with other peer'
+        b'performance during exchanges with other repositories.'
     )
 
-    touches_filelogs = True
-    touches_manifests = True
-    touches_changelog = False
+    touches_filelogs: ClassVar[bool] = True
+    touches_manifests: ClassVar[bool] = True
+    touches_changelog: ClassVar[bool] = False
 
 
 @registerformatvariant
 class persistentnodemap(requirementformatvariant):
-    name = b'persistent-nodemap'
+    name: ClassVar[bytes] = b'persistent-nodemap'
 
-    _requirement = requirements.NODEMAP_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.NODEMAP_REQUIREMENT
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'persist the node -> rev mapping on disk to speedup lookup'
     )
 
-    upgrademessage = _(b'Speedup revision lookup by node id.')
+    upgrademessage: ClassVar[bytes] = _(b'Speedup revision lookup by node id.')
 
 
 @registerformatvariant
 class copiessdc(requirementformatvariant):
-    name = b'copies-sdc'
+    name: ClassVar[bytes] = b'copies-sdc'
 
-    _requirement = requirements.COPIESSDC_REQUIREMENT
+    _requirement: ClassVar[RequirementT] = requirements.COPIESSDC_REQUIREMENT
 
-    default = False
+    default: ClassVar[bool] = False
 
-    description = _(b'Stores copies information alongside changesets.')
+    description: ClassVar[bytes] = _(
+        b'Stores copies information alongside changesets.'
+    )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'Allows to use more efficient algorithm to deal with copy tracing.'
     )
 
-    touches_filelogs = False
-    touches_manifests = False
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
 
 
 @registerformatvariant
 class revlogv2(requirementformatvariant):
-    name = b'revlog-v2'
-    _requirement = requirements.REVLOGV2_REQUIREMENT
-    default = False
-    description = _(b'Version 2 of the revlog.')
-    upgrademessage = _(b'very experimental')
+    name: ClassVar[bytes] = b'revlog-v2'
+    _requirement: ClassVar[RequirementT] = requirements.REVLOGV2_REQUIREMENT
+    default: ClassVar[bool] = False
+    description: ClassVar[bytes] = _(b'Version 2 of the revlog.')
+    upgrademessage: ClassVar[bytes] = _(b'very experimental')
 
 
 @registerformatvariant
 class changelogv2(requirementformatvariant):
-    name = b'changelog-v2'
-    _requirement = requirements.CHANGELOGV2_REQUIREMENT
-    default = False
-    description = _(b'An iteration of the revlog focussed on changelog needs.')
-    upgrademessage = _(b'quite experimental')
+    name: ClassVar[bytes] = b'changelog-v2'
+    _requirement: ClassVar[RequirementT] = requirements.CHANGELOGV2_REQUIREMENT
+    default: ClassVar[bool] = False
+    description: ClassVar[bytes] = _(
+        b'An iteration of the revlog focussed on changelog needs.'
+    )
+    upgrademessage: ClassVar[bytes] = _(b'quite experimental')
 
-    touches_filelogs = False
-    touches_manifests = False
+    touches_filelogs: ClassVar[bool] = False
+    touches_manifests: ClassVar[bool] = False
 
 
 @registerformatvariant
 class removecldeltachain(formatvariant):
-    name = b'plain-cl-delta'
+    name: ClassVar[bytes] = b'plain-cl-delta'
 
-    default = True
+    default: ClassVar[bool] = True
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'changelog storage is using deltas instead of '
         b'raw entries; changelog reading and any '
         b'operation relying on changelog data are slower '
         b'than they could be'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'changelog storage will be reformated to '
         b'store raw entries; changelog reading will be '
         b'faster; changelog size may be reduced'
     )
 
     @staticmethod
-    def fromrepo(repo):
+    def fromrepo(repo: RepoT) -> bool:
         # Mercurial 4.0 changed changelogs to not use delta chains. Search for
         # changelogs with deltas.
         cl = repo.unfiltered().changelog
@@ -493,11 +524,11 @@ class removecldeltachain(formatvariant):
         return all(rev == chainbase(rev) for rev in some_rev)
 
     @staticmethod
-    def fromconfig(repo):
+    def fromconfig(repo: RepoT) -> bool:
         return True
 
 
-_has_zstd = (
+_has_zstd: Final[bool] = bool(
     b'zstd' in util.compengines
     and util.compengines[b'zstd'].available()
     and util.compengines[b'zstd'].revlogheader()
@@ -506,24 +537,21 @@ _has_zstd = (
 
 @registerformatvariant
 class compressionengine(formatvariant):
-    name = b'compression'
+    name: ClassVar[bytes] = b'compression'
 
-    if _has_zstd:
-        default = b'zstd'
-    else:
-        default = b'zlib'
+    default: ClassVar[bytes] = b'zstd' if _has_zstd else b'zlib'
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'Compresion algorithm used to compress data. '
         b'Some engine are faster than other'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'revlog content will be recompressed with the new algorithm.'
     )
 
     @classmethod
-    def fromrepo(cls, repo):
+    def fromrepo(cls, repo: RepoT) -> bytes:
         # we allow multiple compression engine requirement to co-exist because
         # strickly speaking, revlog seems to support mixed compression style.
         #
@@ -536,7 +564,7 @@ class compressionengine(formatvariant):
         return compression
 
     @classmethod
-    def fromconfig(cls, repo):
+    def fromconfig(cls, repo: RepoT) -> bytes:
         compengines = repo.ui.configlist(b'format', b'revlog-compression')
         # return the first valid value as the selection code would do
         for comp in compengines:
@@ -551,15 +579,15 @@ class compressionengine(formatvariant):
 
 @registerformatvariant
 class compressionlevel(formatvariant):
-    name = b'compression-level'
-    default = b'default'
+    name: ClassVar[bytes] = b'compression-level'
+    default: ClassVar[bytes] = b'default'
 
-    description = _(b'compression level')
+    description: ClassVar[bytes] = _(b'compression level')
 
-    upgrademessage = _(b'revlog content will be recompressed')
+    upgrademessage: ClassVar[bytes] = _(b'revlog content will be recompressed')
 
     @classmethod
-    def fromrepo(cls, repo):
+    def fromrepo(cls, repo: RepoT) -> bytes:
         comp = compressionengine.fromrepo(repo)
         level = None
         if comp == b'zlib':
@@ -571,7 +599,7 @@ class compressionlevel(formatvariant):
         return b"%d" % level
 
     @classmethod
-    def fromconfig(cls, repo):
+    def fromconfig(cls, repo: RepoT) -> bytes:
         comp = compressionengine.fromconfig(repo)
         level = None
         if comp == b'zlib':
@@ -583,7 +611,7 @@ class compressionlevel(formatvariant):
         return b"%d" % level
 
 
-def find_format_upgrades(repo):
+def find_format_upgrades(repo: RepoT):
     """returns a list of format upgrades which can be perform on the repo"""
     upgrades = []
 
@@ -598,7 +626,7 @@ def find_format_upgrades(repo):
     return upgrades
 
 
-def find_format_downgrades(repo):
+def find_format_downgrades(repo: RepoT):
     """returns a list of format downgrades which will be performed on the repo
     because of disabled config option for them"""
 
@@ -631,16 +659,36 @@ def register_optimization(obj):
 class optimization(improvement):
     """an improvement subclass dedicated to optimizations"""
 
-    type = OPTIMISATION
+    type: ClassVar[bytes] = OPTIMISATION
+
+
+@register_optimization
+class ReDeltaQuick(optimization):
+    name: ClassVar[bytes] = b're-delta-quick'
+
+    type: ClassVar[bytes] = OPTIMISATION
+
+    description: ClassVar[bytes] = _(
+        b'deltas within internal storage will be reprocessed with minimal '
+        b'overhead, deltas already in store will be reused as often as '
+        b'possible. This provides a way to smooth out potential problematic '
+        b'deltas while avoiding deeper but slower optimization that might '
+        b'not be necessary'
+    )
+
+    upgrademessage: ClassVar[bytes] = _(
+        b"deltas within internal storage will choose a new "
+        b"base revision when the previous one wasn't suitable"
+    )
 
 
 @register_optimization
 class redeltaparents(optimization):
-    name = b're-delta-parent'
+    name: ClassVar[bytes] = b're-delta-parent'
 
-    type = OPTIMISATION
+    type: ClassVar[bytes] = OPTIMISATION
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'deltas within internal storage will be recalculated to '
         b'choose an optimal base revision where this was not '
         b'already done; the size of the repository may shrink and '
@@ -650,7 +698,7 @@ class redeltaparents(optimization):
         b'not run noticeably slower'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'deltas within internal storage will choose a new '
         b'base revision if needed'
     )
@@ -658,11 +706,11 @@ class redeltaparents(optimization):
 
 @register_optimization
 class redeltamultibase(optimization):
-    name = b're-delta-multibase'
+    name: ClassVar[bytes] = b're-delta-multibase'
 
-    type = OPTIMISATION
+    type: ClassVar[bytes] = OPTIMISATION
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'deltas within internal storage will be recalculated '
         b'against multiple base revision and the smallest '
         b'difference will be used; the size of the repository may '
@@ -674,7 +722,7 @@ class redeltamultibase(optimization):
         b'files and thousands of merges'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'deltas within internal storage will choose an '
         b'optimal delta by computing deltas against multiple '
         b'parents; may slow down execution time '
@@ -684,18 +732,18 @@ class redeltamultibase(optimization):
 
 @register_optimization
 class redeltaall(optimization):
-    name = b're-delta-all'
+    name: ClassVar[bytes] = b're-delta-all'
 
-    type = OPTIMISATION
+    type: ClassVar[bytes] = OPTIMISATION
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'deltas within internal storage will always be '
         b'recalculated without reusing prior deltas; this will '
         b'likely make execution run several times slower; this '
         b'optimization is typically not needed'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'deltas within internal storage will be fully '
         b'recomputed; this will likely drastically slow down '
         b'execution time'
@@ -704,11 +752,11 @@ class redeltaall(optimization):
 
 @register_optimization
 class redeltafulladd(optimization):
-    name = b're-delta-fulladd'
+    name: ClassVar[bytes] = b're-delta-fulladd'
 
-    type = OPTIMISATION
+    type: ClassVar[bytes] = OPTIMISATION
 
-    description = _(
+    description: ClassVar[bytes] = _(
         b'every revision will be re-added as if it was new '
         b'content. It will go through the full storage '
         b'mechanism giving extensions a chance to process it '
@@ -716,7 +764,7 @@ class redeltafulladd(optimization):
         b'slower since more logic is involved.'
     )
 
-    upgrademessage = _(
+    upgrademessage: ClassVar[bytes] = _(
         b'each revision will be added as new content to the '
         b'internal storage; this will likely drastically slow '
         b'down execution time, but some extensions might need '
@@ -724,7 +772,9 @@ class redeltafulladd(optimization):
     )
 
 
-def findoptimizations(repo):
+# TODO: UpgradeOperation lacks a repo object, and passes None here.  Either fix
+#  that or drop the repo arg here?
+def findoptimizations(repo: RepoT | None):
     """Determine optimisation that could be used during upgrade"""
     # These are unconditionally added. There is logic later that figures out
     # which ones to apply.
@@ -732,7 +782,11 @@ def findoptimizations(repo):
 
 
 def determine_upgrade_actions(
-    repo, format_upgrades, optimizations, sourcereqs, destreqs
+    repo: RepoT,
+    format_upgrades,
+    optimizations,
+    sourcereqs: RequirementSetT,
+    destreqs: RequirementSetT,
 ):
     """Determine upgrade actions that will be performed.
 
@@ -780,9 +834,11 @@ class BaseOperation:
     UpgradeOperation extends)
     """
 
+    new_requirements: RequirementSetT
+
     def __init__(
         self,
-        new_requirements,
+        new_requirements: RequirementSetT,
         backup_store,
     ):
         self.new_requirements = new_requirements
@@ -795,9 +851,9 @@ class UpgradeOperation(BaseOperation):
 
     def __init__(
         self,
-        ui,
-        new_requirements,
-        current_requirements,
+        ui: UiT,
+        new_requirements: RequirementSetT,
+        current_requirements: RequirementSetT,
         upgrade_actions,
         removed_actions,
         revlogs_to_process,
@@ -843,6 +899,8 @@ class UpgradeOperation(BaseOperation):
             self.delta_reuse_mode = revlog.revlog.DELTAREUSESAMEREVS
         elif b're-delta-fulladd' in upgrade_actions_names:
             self.delta_reuse_mode = revlog.revlog.DELTAREUSEFULLADD
+        elif b're-delta-quick' in upgrade_actions_names:
+            self.delta_reuse_mode = revlog.revlog.DELTAREUSEALWAYS
 
         # should this operation force re-delta of both parents
         self.force_re_delta_both_parents = (
@@ -850,11 +908,11 @@ class UpgradeOperation(BaseOperation):
         )
 
     @property
-    def upgrade_actions_names(self):
+    def upgrade_actions_names(self) -> set[bytes]:
         return {a.name for a in self.upgrade_actions}
 
     @property
-    def requirements_only(self):
+    def requirements_only(self) -> bool:
         # does the operation only touches repository requirement
         return (
             self.touches_requirements
@@ -865,7 +923,7 @@ class UpgradeOperation(BaseOperation):
         )
 
     @property
-    def touches_filelogs(self):
+    def touches_filelogs(self) -> bool:
         for a in self.upgrade_actions:
             # in optimisations, we re-process the revlogs again
             if a.type == OPTIMISATION:
@@ -878,7 +936,7 @@ class UpgradeOperation(BaseOperation):
         return False
 
     @property
-    def touches_manifests(self):
+    def touches_manifests(self) -> bool:
         for a in self.upgrade_actions:
             # in optimisations, we re-process the revlogs again
             if a.type == OPTIMISATION:
@@ -891,7 +949,7 @@ class UpgradeOperation(BaseOperation):
         return False
 
     @property
-    def touches_changelog(self):
+    def touches_changelog(self) -> bool:
         for a in self.upgrade_actions:
             # in optimisations, we re-process the revlogs again
             if a.type == OPTIMISATION:
@@ -904,7 +962,7 @@ class UpgradeOperation(BaseOperation):
         return False
 
     @property
-    def touches_requirements(self):
+    def touches_requirements(self) -> bool | None:  # TODO: fix return paths
         for a in self.upgrade_actions:
             # optimisations are used to re-process revlogs and does not result
             # in a requirement being added or removed
@@ -917,7 +975,7 @@ class UpgradeOperation(BaseOperation):
                 return True
 
     @property
-    def touches_dirstate(self):
+    def touches_dirstate(self) -> bool:
         for a in self.upgrade_actions:
             # revlog optimisations do not affect the dirstate
             if a.type == OPTIMISATION:
@@ -930,7 +988,7 @@ class UpgradeOperation(BaseOperation):
 
         return False
 
-    def _write_labeled(self, l, label: bytes):
+    def _write_labeled(self, l, label: bytes) -> None:
         """
         Utility function to aid writing of a list under one label
         """
@@ -941,7 +999,7 @@ class UpgradeOperation(BaseOperation):
             self.ui.write(r, label=label)
             first = False
 
-    def print_requirements(self):
+    def print_requirements(self) -> None:
         self.ui.write(_(b'requirements\n'))
         self.ui.write(_(b'   preserved: '))
         self._write_labeled(
@@ -962,7 +1020,7 @@ class UpgradeOperation(BaseOperation):
             self.ui.writenoi18n(b'\n')
         self.ui.write(b'\n')
 
-    def print_optimisations(self):
+    def print_optimisations(self) -> None:
         optimisations = [
             a for a in self.upgrade_actions if a.type == OPTIMISATION
         ]
@@ -975,11 +1033,11 @@ class UpgradeOperation(BaseOperation):
             )
             self.ui.write(b'\n\n')
 
-    def print_upgrade_actions(self):
+    def print_upgrade_actions(self) -> None:
         for a in self.upgrade_actions:
             self.ui.status(b'%s\n   %s\n\n' % (a.name, a.upgrademessage))
 
-    def print_affected_revlogs(self):
+    def print_affected_revlogs(self) -> None:
         if not self.revlogs_to_process:
             self.ui.writenoi18n(b'no revlogs to process\n')
         else:
@@ -988,15 +1046,15 @@ class UpgradeOperation(BaseOperation):
                 self.ui.writenoi18n(b'  - %s\n' % r)
         self.ui.writenoi18n(b'\n')
 
-    def print_unused_optimizations(self):
+    def print_unused_optimizations(self) -> None:
         for i in self.unused_optimizations:
             self.ui.status(_(b'%s\n   %s\n\n') % (i.name, i.description))
 
-    def has_upgrade_action(self, name):
+    def has_upgrade_action(self, name: bytes) -> bool:
         """Check whether the upgrade operation will perform this action"""
         return name in self.upgrade_actions_names
 
-    def print_post_op_messages(self):
+    def print_post_op_messages(self) -> None:
         """print post upgrade operation warning messages"""
         for a in self.upgrade_actions:
             if a.postupgrademessage is not None:
@@ -1009,7 +1067,7 @@ class UpgradeOperation(BaseOperation):
 ###  Code checking if a repository can got through the upgrade process at all. #
 
 
-def requiredsourcerequirements(repo):
+def requiredsourcerequirements(repo: RepoT) -> RequirementSetT:
     """Obtain requirements required to be present to upgrade a repo.
 
     An upgrade will not be allowed if the repository doesn't have the
@@ -1021,7 +1079,7 @@ def requiredsourcerequirements(repo):
     }
 
 
-def blocksourcerequirements(repo):
+def blocksourcerequirements(repo: RepoT) -> RequirementSetT:
     """Obtain requirements that will prevent an upgrade from occurring.
 
     An upgrade cannot be performed if the source repository contains a
@@ -1034,7 +1092,7 @@ def blocksourcerequirements(repo):
     }
 
 
-def check_revlog_version(reqs):
+def check_revlog_version(reqs: RequirementSetT) -> None:
     """Check that the requirements contain at least one Revlog version"""
     all_revlogs = {
         requirements.REVLOGV1_REQUIREMENT,
@@ -1045,7 +1103,7 @@ def check_revlog_version(reqs):
         raise error.Abort(msg)
 
 
-def check_source_requirements(repo):
+def check_source_requirements(repo: RepoT) -> None:
     """Ensure that no existing requirements prevent the repository upgrade"""
 
     check_revlog_version(repo.requirements)
@@ -1077,7 +1135,9 @@ def check_source_requirements(repo):
 ### Verify the validity of the planned requirement changes ####################
 
 
-def supportremovedrequirements(repo):
+def supportremovedrequirements(
+    repo: RepoT, new_reqs: RequirementSetT
+) -> RequirementSetT:
     """Obtain requirements that can be removed during an upgrade.
 
     If an upgrade were to create a repository that dropped a requirement,
@@ -1091,7 +1151,6 @@ def supportremovedrequirements(repo):
         requirements.DIRSTATE_TRACKED_HINT_V1,
         requirements.DIRSTATE_V2_REQUIREMENT,
         requirements.DOTENCODE_REQUIREMENT,
-        requirements.FILELOG_METAFLAG_REQUIREMENT,
         requirements.NODEMAP_REQUIREMENT,
         requirements.PLAIN_ENCODE_REQUIREMENT,
         requirements.REVLOGV1_REQUIREMENT,
@@ -1099,6 +1158,10 @@ def supportremovedrequirements(repo):
         requirements.SHARESAFE_REQUIREMENT,
         requirements.SPARSEREVLOG_REQUIREMENT,
     }
+    if requirements.FILEINDEXV1_REQUIREMENT in new_reqs:
+        supported.add(requirements.FNCACHE_REQUIREMENT)
+    elif requirements.FNCACHE_REQUIREMENT in new_reqs:
+        supported.add(requirements.FILEINDEXV1_REQUIREMENT)
     for name in compression.compengines:
         engine = compression.compengines[name]
         if engine.available() and engine.revlogheader():
@@ -1108,7 +1171,7 @@ def supportremovedrequirements(repo):
     return supported
 
 
-def supporteddestrequirements(repo):
+def supporteddestrequirements(repo: RepoT) -> RequirementSetT:
     """Obtain requirements that upgrade supports in the destination.
 
     If the result of the upgrade would have requirements not in this set,
@@ -1123,7 +1186,7 @@ def supporteddestrequirements(repo):
         requirements.DIRSTATE_TRACKED_HINT_V1,
         requirements.DIRSTATE_V2_REQUIREMENT,
         requirements.DOTENCODE_REQUIREMENT,
-        requirements.FILELOG_METAFLAG_REQUIREMENT,
+        requirements.FILEINDEXV1_REQUIREMENT,
         requirements.FNCACHE_REQUIREMENT,
         requirements.GENERALDELTA_REQUIREMENT,
         requirements.NARROW_REQUIREMENT,
@@ -1146,7 +1209,7 @@ def supporteddestrequirements(repo):
     return supported
 
 
-def allowednewrequirements(repo):
+def allowednewrequirements(repo: RepoT) -> RequirementSetT:
     """Obtain requirements that can be added to a repository during upgrade.
 
     This is used to disallow proposed requirements from being added when
@@ -1163,7 +1226,7 @@ def allowednewrequirements(repo):
         requirements.DIRSTATE_TRACKED_HINT_V1,
         requirements.DIRSTATE_V2_REQUIREMENT,
         requirements.DOTENCODE_REQUIREMENT,
-        requirements.FILELOG_METAFLAG_REQUIREMENT,
+        requirements.FILEINDEXV1_REQUIREMENT,
         requirements.FNCACHE_REQUIREMENT,
         requirements.GENERALDELTA_REQUIREMENT,
         requirements.NODEMAP_REQUIREMENT,
@@ -1182,10 +1245,10 @@ def allowednewrequirements(repo):
     return supported
 
 
-def check_requirements_changes(repo, new_reqs):
+def check_requirements_changes(repo: RepoT, new_reqs: RequirementSetT) -> None:
     old_reqs = repo.requirements
     check_revlog_version(repo.requirements)
-    support_removal = supportremovedrequirements(repo)
+    support_removal = supportremovedrequirements(repo, new_reqs)
     no_remove_reqs = old_reqs - new_reqs - support_removal
     if no_remove_reqs:
         msg = _(b'cannot upgrade repository; requirement would be removed: %s')
